import logging
import re
from typing import Any, TYPE_CHECKING
from collections.abc import Callable, Sequence

from slither.core.declarations import (
    Modifier,
    EventContract,
    EnumContract,
    StructureContract,
    Function,
)
from slither.core.declarations.contract import Contract
from slither.core.declarations.custom_error_contract import CustomErrorContract
from slither.core.declarations.function_contract import FunctionContract
from slither.core.solidity_types import ElementaryType, TypeAliasContract
from slither.core.variables.state_variable import StateVariable
from slither.core.expressions.expression import Expression
from slither.solc_parsing.declarations.caller_context import CallerContextExpression
from slither.solc_parsing.declarations.custom_error import CustomErrorSolc
from slither.solc_parsing.declarations.event_contract import EventContractSolc
from slither.solc_parsing.declarations.function import FunctionSolc
from slither.solc_parsing.declarations.modifier import ModifierSolc
from slither.solc_parsing.declarations.structure_contract import StructureContractSolc
from slither.solc_parsing.exceptions import ParsingError, VariableNotFound
from slither.solc_parsing.solidity_types.type_parsing import parse_type
from slither.solc_parsing.variables.state_variable import StateVariableSolc
from slither.solc_parsing.expressions.expression_parsing import parse_expression
from slither.utils.using_for import USING_FOR_KEY
from slither.visitors.expression.constants_folding import ConstantFolding, NotConstant

LOGGER = logging.getLogger("ContractSolcParsing")

if TYPE_CHECKING:
    from slither.solc_parsing.slither_compilation_unit_solc import SlitherCompilationUnitSolc
    from slither.core.compilation_unit import SlitherCompilationUnit


class ContractSolc(CallerContextExpression):
    def __init__(
        self, slither_parser: "SlitherCompilationUnitSolc", contract: Contract, data: dict[str, Any]
    ) -> None:
        # assert slitherSolc.solc_version.startswith('0.4')

        self._contract = contract
        self._slither_parser = slither_parser
        self._data = data

        self._functionsNotParsed: list[dict] = []
        self._modifiersNotParsed: list[dict] = []
        self._functions_no_params: list[FunctionSolc] = []
        self._modifiers_no_params: list[ModifierSolc] = []
        self._eventsNotParsed: list[dict] = []
        self._variablesNotParsed: list[dict] = []
        self._enumsNotParsed: list[dict] = []
        self._structuresNotParsed: list[dict] = []
        self._usingForNotParsed: list[dict] = []
        self._customErrorsNotParsed: list[dict] = []

        self._functions_parser: list[FunctionSolc] = []
        self._modifiers_parser: list[ModifierSolc] = []
        self._structures_parser: list[StructureContractSolc] = []
        self._custom_errors_parser: list[CustomErrorSolc] = []

        self._storage_layout_parsed_expression: Expression | None = None

        self._is_analyzed: bool = False

        # use to remap inheritance id
        self._remapping: dict[str, str] = {}

        # (referencedDeclaration, offset)
        self.baseContracts: list[tuple[int, str]] = []
        self.baseConstructorContractsCalled: list[str] = []
        self._linearized_base_contracts: list[int]

        self._variables_parser: list[StateVariableSolc] = []

        # Export info
        if self.is_compact_ast:
            self._contract.name = self._data["name"]
            self._handle_comment(self._data)
        else:
            self._contract.name = self._data["attributes"][self.get_key()]
            self._handle_comment(self._data["attributes"])

        self._contract.id = self._data["id"]

        self._parse_contract_info()
        self._parse_contract_items()

    ###################################################################################
    ###################################################################################
    # region General Properties
    ###################################################################################
    ###################################################################################

    @property
    def is_analyzed(self) -> bool:
        return self._is_analyzed

    def set_is_analyzed(self, is_analyzed: bool) -> None:
        self._is_analyzed = is_analyzed

    @property
    def underlying_contract(self) -> Contract:
        return self._contract

    @property
    def linearized_base_contracts(self) -> list[int]:
        return self._linearized_base_contracts

    @property
    def compilation_unit(self) -> "SlitherCompilationUnit":
        return self._contract.compilation_unit

    @property
    def slither_parser(self) -> "SlitherCompilationUnitSolc":
        return self._slither_parser

    @property
    def functions_parser(self) -> list["FunctionSolc"]:
        return self._functions_parser

    @property
    def modifiers_parser(self) -> list["ModifierSolc"]:
        return self._modifiers_parser

    @property
    def structures_not_parsed(self) -> list[dict]:
        return self._structuresNotParsed

    @property
    def enums_not_parsed(self) -> list[dict]:
        return self._enumsNotParsed

    ###################################################################################
    ###################################################################################
    # region AST
    ###################################################################################
    ###################################################################################

    def get_key(self) -> str:
        return self._slither_parser.get_key()

    def get_children(self, key: str = "nodes") -> str:
        if self.is_compact_ast:
            return key
        return "children"

    @property
    def remapping(self) -> dict[str, str]:
        return self._remapping

    @property
    def is_compact_ast(self) -> bool:
        return self._slither_parser.is_compact_ast

    # endregion
    ###################################################################################
    ###################################################################################
    # region SlithIR
    ###################################################################################
    ###################################################################################

    def _parse_contract_info(self) -> None:
        if self.is_compact_ast:
            attributes = self._data
        else:
            attributes = self._data["attributes"]

        self._contract.is_interface = False
        if "contractKind" in attributes:
            if attributes["contractKind"] == "interface":
                self._contract.is_interface = True
            elif attributes["contractKind"] == "library":
                self._contract.is_library = True
            self._contract.contract_kind = attributes["contractKind"]

        self._contract.is_fully_implemented = attributes["fullyImplemented"]
        self._linearized_base_contracts = attributes["linearizedBaseContracts"]

        if "storageLayout" in attributes:
            # For now we care only about the actual value, hence we immediately parse the expression
            # and ConstantFold it later on since it could be using a TopLevel variable
            self._storage_layout_parsed_expression = parse_expression(
                attributes["storageLayout"]["baseSlotExpression"], self
            )

        if "abstract" in attributes:
            self._contract.is_abstract = attributes["abstract"]

        # Parse base contract information
        self._parse_base_contract_info()

        # trufle does some re-mapping of id
        if "baseContracts" in self._data:
            for elem in self._data["baseContracts"]:
                if elem["nodeType"] == "InheritanceSpecifier":
                    self._remapping[elem["baseName"]["referencedDeclaration"]] = elem["baseName"][
                        "name"
                    ]

    def _parse_base_contract_info(self) -> None:
        # Parse base contracts (immediate, non-linearized)
        if self.is_compact_ast:
            # Parse base contracts + constructors in compact-ast
            if "baseContracts" in self._data:
                for base_contract in self._data["baseContracts"]:
                    if base_contract["nodeType"] != "InheritanceSpecifier":
                        continue
                    if (
                        "baseName" not in base_contract
                        or "referencedDeclaration" not in base_contract["baseName"]
                    ):
                        continue

                    # Obtain our contract reference and add it to our base contract list
                    referencedDeclaration = base_contract["baseName"]["referencedDeclaration"]
                    self.baseContracts.append(
                        (referencedDeclaration, base_contract["baseName"]["src"])
                    )

                    # If we have defined arguments in our arguments object, this is a constructor invocation.
                    # (note: 'arguments' can be [], which is not the same as None. [] implies a constructor was
                    #  called with no arguments, while None implies no constructor was called).
                    if "arguments" in base_contract and base_contract["arguments"] is not None:
                        self.baseConstructorContractsCalled.append(referencedDeclaration)
        else:
            # Parse base contracts + constructors in legacy-ast
            if "children" in self._data:
                for base_contract in self._data["children"]:
                    if base_contract["name"] != "InheritanceSpecifier":
                        continue
                    if "children" not in base_contract or len(base_contract["children"]) == 0:
                        continue
                    # Obtain all items for this base contract specification (base contract, followed by arguments)
                    base_contract_items = base_contract["children"]
                    if (
                        "name" not in base_contract_items[0]
                        or base_contract_items[0]["name"] != "UserDefinedTypeName"
                    ):
                        continue
                    if (
                        "attributes" not in base_contract_items[0]
                        or "referencedDeclaration" not in base_contract_items[0]["attributes"]
                    ):
                        continue

                    # Obtain our contract reference and add it to our base contract list
                    referencedDeclaration = base_contract_items[0]["attributes"][
                        "referencedDeclaration"
                    ]

                    self.baseContracts.append(
                        (referencedDeclaration, base_contract_items[0]["src"])
                    )

                    # If we have an 'attributes'->'arguments' which is None, this is not a constructor call.
                    if (
                        "attributes" not in base_contract
                        or "arguments" not in base_contract["attributes"]
                        or base_contract["attributes"]["arguments"] is not None
                    ):
                        self.baseConstructorContractsCalled.append(referencedDeclaration)

    def _parse_contract_items(self) -> None:
        if self.get_children() not in self._data:  # empty contract
            return
        for item in self._data[self.get_children()]:
            if item[self.get_key()] == "FunctionDefinition":
                self._functionsNotParsed.append(item)
            elif item[self.get_key()] == "EventDefinition":
                self._eventsNotParsed.append(item)
            elif item[self.get_key()] == "InheritanceSpecifier":
                # we dont need to parse it as it is redundant
                # with self.linearizedBaseContracts
                continue
            elif item[self.get_key()] == "VariableDeclaration":
                self._variablesNotParsed.append(item)
            elif item[self.get_key()] == "EnumDefinition":
                self._enumsNotParsed.append(item)
            elif item[self.get_key()] == "ModifierDefinition":
                self._modifiersNotParsed.append(item)
            elif item[self.get_key()] == "StructDefinition":
                self._structuresNotParsed.append(item)
            elif item[self.get_key()] == "UsingForDirective":
                self._usingForNotParsed.append(item)
            elif item[self.get_key()] == "ErrorDefinition":
                self._customErrorsNotParsed.append(item)
            elif item[self.get_key()] == "UserDefinedValueTypeDefinition":
                self._parse_type_alias(item)
            else:
                raise ParsingError("Unknown contract item: " + item[self.get_key()])
        return

    def parse_type_alias(self) -> None:
        # We keep parse_ in the name just to keep the naming convention even if we already parsed them initially.
        # Here we only update the current contract type_aliases_as_dict with the fathers' values
        # It's useful to keep using the same pattern anyway as we know all the fathers have been analyzed
        for father in self._contract.inheritance_reverse:
            self._contract.type_aliases_as_dict.update(father.type_aliases_as_dict)

    def _parse_type_alias(self, item: dict) -> None:
        assert "name" in item
        assert "underlyingType" in item
        underlying_type = item["underlyingType"]
        assert "nodeType" in underlying_type and underlying_type["nodeType"] == "ElementaryTypeName"
        assert "name" in underlying_type

        original_type = ElementaryType(underlying_type["name"])

        # For user defined types defined at the contract level the lookup can be done
        # Using the name or the canonical name
        # For example during the type parsing the canonical name
        # Note that Solidity allows shadowing of user defined types
        # Between top level and contract definitions
        alias = item["name"]
        alias_canonical = self._contract.name + "." + item["name"]

        type_alias = TypeAliasContract(original_type, alias, self.underlying_contract)
        type_alias.set_offset(item["src"], self.compilation_unit)
        self._contract.type_aliases_as_dict[alias] = type_alias
        self._contract.file_scope.type_aliases[alias_canonical] = type_alias

    def _parse_struct(self, struct: dict) -> None:
        st = StructureContract(self._contract.compilation_unit)
        st.set_contract(self._contract)
        st.set_offset(struct["src"], self._contract.compilation_unit)

        st_parser = StructureContractSolc(st, struct, self)  # type: ignore
        self._contract.structures_as_dict[st.name] = st
        self._structures_parser.append(st_parser)

    def parse_structs(self) -> None:
        for father in self._contract.inheritance_reverse:
            self._contract.structures_as_dict.update(father.structures_as_dict)

        for struct in self._structuresNotParsed:
            self._parse_struct(struct)
        self._structuresNotParsed = []

    def _parse_custom_error(self, custom_error: dict) -> None:
        ce = CustomErrorContract(self.compilation_unit)
        ce.set_contract(self._contract)
        ce.set_offset(custom_error["src"], self.compilation_unit)

        ce_parser = CustomErrorSolc(ce, custom_error, self, self._slither_parser)
        self._contract.custom_errors_as_dict[ce.name] = ce
        self._custom_errors_parser.append(ce_parser)

    def parse_custom_errors(self) -> None:
        for father in self._contract.inheritance_reverse:
            self._contract.custom_errors_as_dict.update(father.custom_errors_as_dict)

        for custom_error in self._customErrorsNotParsed:
            self._parse_custom_error(custom_error)
        self._customErrorsNotParsed = []

    def parse_state_variables(self) -> None:
        for father in self._contract.inheritance_reverse:
            self._contract.variables_as_dict.update(
                {
                    name: v
                    for name, v in father.variables_as_dict.items()
                    if v.visibility != "private"
                }
            )
            self._contract.add_state_variables_ordered(
                [
                    var
                    for var in father.state_variables_ordered
                    if var not in self._contract.state_variables_ordered
                ]
            )

        for varNotParsed in self._variablesNotParsed:
            var = StateVariable()
            var.set_offset(varNotParsed["src"], self._contract.compilation_unit)
            var.set_contract(self._contract)

            var_parser = StateVariableSolc(var, varNotParsed)
            self._variables_parser.append(var_parser)

            assert var.name
            if var_parser.reference_id is not None:
                self._contract.state_variables_by_ref_id[var_parser.reference_id] = var
            self._contract.variables_as_dict[var.name] = var
            self._contract.add_state_variables_ordered([var])

    def _parse_modifier(self, modifier_data: dict) -> None:
        modif = Modifier(self._contract.compilation_unit)
        modif.set_offset(modifier_data["src"], self._contract.compilation_unit)
        modif.set_contract(self._contract)
        modif.set_contract_declarer(self._contract)

        modif_parser = ModifierSolc(modif, modifier_data, self, self.slither_parser)  # type: ignore
        self._contract.compilation_unit.add_modifier(modif)
        self._modifiers_no_params.append(modif_parser)
        self._modifiers_parser.append(modif_parser)

        self._slither_parser.add_function_or_modifier_parser(modif_parser)

    def parse_modifiers(self) -> None:
        for modifier in self._modifiersNotParsed:
            self._parse_modifier(modifier)
        self._modifiersNotParsed = []

    def _parse_function(self, function_data: dict) -> None:
        func = FunctionContract(self._contract.compilation_unit)
        func.set_offset(function_data["src"], self._contract.compilation_unit)
        func.set_contract(self._contract)
        func.set_contract_declarer(self._contract)

        func_parser = FunctionSolc(func, function_data, self, self._slither_parser)  # type: ignore
        self._contract.compilation_unit.add_function(func)
        self._functions_no_params.append(func_parser)
        self._functions_parser.append(func_parser)

        self._slither_parser.add_function_or_modifier_parser(func_parser)

    def parse_functions(self) -> None:
        for function in self._functionsNotParsed:
            self._parse_function(function)

        self._functionsNotParsed = []

    # endregion
    ###################################################################################
    ###################################################################################
    # region Analyze
    ###################################################################################
    ###################################################################################

    def log_incorrect_parsing(self, error: str) -> None:
        if self._contract.compilation_unit.core.disallow_partial:
            raise ParsingError(error)
        LOGGER.error(error)
        self._contract.is_incorrectly_constructed = True

    def analyze_storage_layout(self) -> None:
        if self._storage_layout_parsed_expression is not None:
            try:
                self._contract.custom_storage_layout = (
                    ConstantFolding(self._storage_layout_parsed_expression, "uint256")
                    .result()
                    .value
                )
            except NotConstant as e:
                self.log_incorrect_parsing(
                    f"Error when folding the custom storage layout value {e}"
                )

    def analyze_content_modifiers(self) -> None:
        try:
            for modifier_parser in self._modifiers_parser:
                modifier_parser.analyze_content()
        except (VariableNotFound, KeyError) as e:
            self.log_incorrect_parsing(f"Missing modifier {e}")

    def analyze_content_functions(self) -> None:
        try:
            for function_parser in self._functions_parser:
                function_parser.analyze_content()
        except (VariableNotFound, KeyError, ParsingError) as e:
            self.log_incorrect_parsing(f"Missing function {e}")

    def analyze_params_modifiers(self) -> None:
        try:
            elements_no_params = self._modifiers_no_params
            getter = lambda c: c.modifiers_parser
            getter_available = lambda c: c.modifiers_declared
            Cls = Modifier
            Cls_parser = ModifierSolc
            modifiers = self._analyze_params_elements(
                elements_no_params,
                getter,
                getter_available,
                Cls,
                Cls_parser,
                self._modifiers_parser,
            )
            # modifiers will be using Modifier so we can ignore the next type check
            self._contract.set_modifiers(modifiers)  # type: ignore
        except (VariableNotFound, KeyError) as e:
            self.log_incorrect_parsing(f"Missing params {e}")
        self._modifiers_no_params = []

    def analyze_params_functions(self) -> None:
        try:
            elements_no_params = self._functions_no_params
            getter = lambda c: c.functions_parser
            getter_available = lambda c: c.functions_declared
            Cls = FunctionContract
            Cls_parser = FunctionSolc
            functions = self._analyze_params_elements(
                elements_no_params,
                getter,
                getter_available,
                Cls,
                Cls_parser,
                self._functions_parser,
            )
            # function will be using FunctionContract so we can ignore the next type check
            self._contract.set_functions(functions)  # type: ignore
        except (VariableNotFound, KeyError) as e:
            self.log_incorrect_parsing(f"Missing params {e}")
        self._functions_no_params = []

    def _analyze_params_element(
        self,
        Cls: Callable,
        Cls_parser: Callable,
        element_parser: FunctionSolc,
        explored_reference_id: set[str],
        parser: list[FunctionSolc] | list[ModifierSolc],
        all_elements: dict[str, Function],
    ) -> None:
        elem = Cls(self._contract.compilation_unit)
        elem.set_contract(self._contract)
        underlying_function = element_parser.underlying_function
        # TopLevel function are not analyzed here
        assert isinstance(underlying_function, FunctionContract)
        elem.set_contract_declarer(underlying_function.contract_declarer)
        elem.set_offset(
            element_parser.function_not_parsed["src"],
            self._contract.compilation_unit,
        )

        elem_parser = Cls_parser(
            elem, element_parser.function_not_parsed, self, self.slither_parser
        )
        if (
            element_parser.underlying_function.id
            and element_parser.underlying_function.id in explored_reference_id
        ):
            # Already added from other fathers
            return
        if element_parser.underlying_function.id:
            explored_reference_id.add(element_parser.underlying_function.id)
        elem_parser.analyze_params()
        if isinstance(elem, Modifier):
            self._contract.compilation_unit.add_modifier(elem)
        else:
            self._contract.compilation_unit.add_function(elem)

        self._slither_parser.add_function_or_modifier_parser(elem_parser)

        all_elements[elem.canonical_name] = elem
        parser.append(elem_parser)

    def _analyze_params_elements(
        self,
        elements_no_params: Sequence[FunctionSolc],
        getter: Callable[["ContractSolc"], list[FunctionSolc]],
        getter_available: Callable[[Contract], list[FunctionContract]],
        Cls: Callable,
        Cls_parser: Callable,
        parser: list[FunctionSolc] | list[ModifierSolc],
    ) -> dict[str, Function]:
        """
        Analyze the parameters of the given elements (Function or Modifier).
        The function iterates over the inheritance to create an instance or inherited elements (Function or Modifier)
        If the element is shadowed, set is_shadowed to True

        :param elements_no_params: list of elements to analyzer
        :param getter: fun x
        :param getter_available: fun x
        :param Cls: Class to create for collision
        :return:
        """
        all_elements: dict[str, Function] = {}

        explored_reference_id: set[str] = set()
        try:
            for father in self._contract.inheritance:
                father_parser = self._slither_parser.underlying_contract_to_parser[father]
                for element_parser in getter(father_parser):  # type: ignore
                    self._analyze_params_element(
                        Cls, Cls_parser, element_parser, explored_reference_id, parser, all_elements
                    )

            accessible_elements = self._contract.available_elements_from_inheritances(
                all_elements, getter_available
            )

            # If there is a constructor in the functions
            # We remove the previous constructor
            # As only one constructor is present per contracts
            #
            # Note: contract.all_functions_called returns the constructors of the base contracts
            has_constructor = False
            for element_parser in elements_no_params:
                element_parser.analyze_params()
                if element_parser.underlying_function.is_constructor:
                    has_constructor = True

            if has_constructor:
                _accessible_functions = {
                    k: v for (k, v) in accessible_elements.items() if not v.is_constructor
                }

            for element_parser in elements_no_params:
                accessible_elements[element_parser.underlying_function.full_name] = (
                    element_parser.underlying_function
                )
                all_elements[element_parser.underlying_function.canonical_name] = (
                    element_parser.underlying_function
                )

            for element in all_elements.values():
                if accessible_elements[element.full_name] != all_elements[element.canonical_name]:
                    element.is_shadowed = True
                    accessible_elements[element.full_name].shadows = True
        except (VariableNotFound, KeyError) as e:
            self.log_incorrect_parsing(
                f"Missing params {e} {self._contract.source_mapping.to_detailed_str()}"
            )
        return all_elements

    def analyze_constant_state_variables(self) -> None:
        for var_parser in self._variables_parser:
            if var_parser.underlying_variable.is_constant:
                # can't parse constant expression based on function calls
                try:
                    var_parser.analyze(self)
                except (VariableNotFound, KeyError) as e:
                    LOGGER.error(e)

    def analyze_state_variables(self) -> None:
        try:
            for var_parser in self._variables_parser:
                var_parser.analyze(self)
            return
        except (VariableNotFound, KeyError) as e:
            self.log_incorrect_parsing(f"Missing state variable {e}")

    def analyze_using_for(self) -> None:
        try:
            for father in self._contract.inheritance:
                self._contract.using_for.update(father.using_for)
            if self.is_compact_ast:
                for using_for in self._usingForNotParsed:
                    if using_for.get("typeName"):
                        type_name: USING_FOR_KEY = parse_type(using_for["typeName"], self)
                    else:
                        type_name = "*"
                    if type_name not in self._contract.using_for:
                        self._contract.using_for[type_name] = []

                    if "libraryName" in using_for:
                        self._contract.using_for[type_name].append(
                            parse_type(using_for["libraryName"], self)
                        )
                    else:
                        # We have a list of functions. A function can be topLevel or a library function
                        self._analyze_function_list(using_for["functionList"], type_name)
            else:
                for using_for in self._usingForNotParsed:
                    children = using_for[self.get_children()]
                    assert children and len(children) <= 2
                    if len(children) == 2:
                        new = parse_type(children[0], self)
                        old: USING_FOR_KEY = parse_type(children[1], self)
                    else:
                        new = parse_type(children[0], self)
                        old = "*"
                    if old not in self._contract.using_for:
                        self._contract.using_for[old] = []
                    self._contract.using_for[old].append(new)
            self._usingForNotParsed = []
        except (VariableNotFound, KeyError) as e:
            self.log_incorrect_parsing(f"Missing using for {e}")

    def _analyze_function_list(self, function_list: list, type_name: USING_FOR_KEY) -> None:
        for f in function_list:
            full_name_split = f["function"]["name"].split(".")
            if len(full_name_split) == 1:
                # Top level function
                function_name = full_name_split[0]
                self._analyze_top_level_function(function_name, type_name)
            elif len(full_name_split) == 2:
                # It can be a top level function behind an aliased import
                # or a library function
                first_part = full_name_split[0]
                function_name = full_name_split[1]
                self._check_aliased_import(first_part, function_name, type_name)
            else:
                # MyImport.MyLib.a we don't care of the alias
                library_name = full_name_split[1]
                function_name = full_name_split[2]
                self._analyze_library_function(library_name, function_name, type_name)

    def _check_aliased_import(
        self, first_part: str, function_name: str, type_name: USING_FOR_KEY
    ) -> None:
        # We check if the first part appear as alias for an import
        # if it is then function_name must be a top level function
        # otherwise it's a library function
        for i in self._contract.file_scope.imports:
            if i.alias == first_part:
                self._analyze_top_level_function(function_name, type_name)
                return
        self._analyze_library_function(first_part, function_name, type_name)

    def _analyze_top_level_function(self, function_name: str, type_name: USING_FOR_KEY) -> None:
        for tl_function in self.compilation_unit.functions_top_level:
            if tl_function.name == function_name:
                self._contract.using_for[type_name].append(tl_function)

    def _analyze_library_function(
        self, library_name: str, function_name: str, type_name: USING_FOR_KEY
    ) -> None:
        # Get the library function
        found = False
        for c in self.compilation_unit.contracts:
            if found:
                break
            if c.name == library_name:
                for f in c.functions:
                    if f.name == function_name:
                        self._contract.using_for[type_name].append(f)
                        found = True
                        break
        if not found:
            self.log_incorrect_parsing(
                f"Contract level using for: Library {library_name} - function {function_name} not found"
            )

    def analyze_enums(self) -> None:
        try:
            for father in self._contract.inheritance:
                self._contract.enums_as_dict.update(father.enums_as_dict)

            for enum in self._enumsNotParsed:
                # for enum, we can parse and analyze it
                # at the same time
                self._analyze_enum(enum)
            self._enumsNotParsed = []
        except (VariableNotFound, KeyError) as e:
            self.log_incorrect_parsing(f"Missing enum {e}")

    def _analyze_enum(
        self,
        enum: dict,
    ) -> None:
        # Enum can be parsed in one pass
        if self.is_compact_ast:
            name = enum["name"]
            canonicalName = enum["canonicalName"]
        else:
            name = enum["attributes"][self.get_key()]
            if "canonicalName" in enum["attributes"]:
                canonicalName = enum["attributes"]["canonicalName"]
            else:
                canonicalName = self._contract.name + "." + name
        values = []
        for child in enum[self.get_children("members")]:
            assert child[self.get_key()] == "EnumValue"
            if self.is_compact_ast:
                values.append(child["name"])
            else:
                values.append(child["attributes"][self.get_key()])

        new_enum = EnumContract(name, canonicalName, values)
        new_enum.set_contract(self._contract)
        new_enum.set_offset(enum["src"], self._contract.compilation_unit)
        self._contract.enums_as_dict[canonicalName] = new_enum

    def _analyze_struct(self, struct: StructureContractSolc) -> None:
        struct.analyze()

    def analyze_structs(self) -> None:
        try:
            for struct in self._structures_parser:
                self._analyze_struct(struct)
        except (VariableNotFound, KeyError) as e:
            self.log_incorrect_parsing(f"Missing struct {e}")

    def analyze_custom_errors(self) -> None:
        for custom_error in self._custom_errors_parser:
            custom_error.analyze_params()

    def analyze_events(self) -> None:
        try:
            for father in self._contract.inheritance_reverse:
                self._contract.events_as_dict.update(father.events_as_dict)

            for event_to_parse in self._eventsNotParsed:
                event = EventContract()
                event.set_contract(self._contract)
                event.set_offset(event_to_parse["src"], self._contract.compilation_unit)

                event_parser = EventContractSolc(event, event_to_parse, self)  # type: ignore
                event_parser.analyze()  # type: ignore
                self._contract.events_as_dict[event.full_name] = event
        except (VariableNotFound, KeyError) as e:
            self.log_incorrect_parsing(f"Missing event {e}")

        self._eventsNotParsed = []

    # endregion
    ###################################################################################
    ###################################################################################
    # region Internal
    ###################################################################################
    ###################################################################################

    def delete_content(self) -> None:
        """
        Remove everything not parsed from the contract
        This is used only if something went wrong with the inheritance parsing
        :return:
        """
        self._functionsNotParsed = []
        self._modifiersNotParsed = []
        self._functions_no_params = []
        self._modifiers_no_params = []
        self._eventsNotParsed = []
        self._variablesNotParsed = []
        self._enumsNotParsed = []
        self._structuresNotParsed = []
        self._usingForNotParsed = []
        self._customErrorsNotParsed = []

    def _handle_comment(self, attributes: dict) -> None:
        """
        Save the contract comment in self.comments
        And handle custom slither comments

        Args:
            attributes:

        Returns:

        """
        # Old solc versions store the comment in attributes["documentation"]
        # More recent ones store it in attributes["documentation"]["text"]
        if (
            "documentation" in attributes
            and attributes["documentation"] is not None
            and (
                "text" in attributes["documentation"]
                or isinstance(attributes["documentation"], str)
            )
        ):
            text = (
                attributes["documentation"]
                if isinstance(attributes["documentation"], str)
                else attributes["documentation"]["text"]
            )
            self._contract.comments = text

            # Look for custom comments
            candidates = text.replace("\n", ",").split(",")

            for candidate in candidates:
                if "@custom:security isDelegatecallProxy" in candidate:
                    self._contract.is_upgradeable_proxy = True
                if "@custom:security isUpgradeable" in candidate:
                    self._contract.is_upgradeable = True

                version_name = re.search(r"@custom:version name=([\w-]+)", candidate)
                if version_name:
                    self._contract.upgradeable_version = version_name.group(1)

    # endregion
    ###################################################################################
    ###################################################################################
    # region Built in definitions
    ###################################################################################
    ###################################################################################

    def __hash__(self) -> int:
        return self._contract.id

    # endregion



from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from slither.core.declarations.contract import Contract

_TESTS_PATTERNS = ["Test", "test", "Mock", "mock"]
TESTS_PATTERNS = _TESTS_PATTERNS + [x + "s" for x in _TESTS_PATTERNS]


def _is_test_pattern(txt: str, pattern: str) -> bool:
    """
    Check if the txt starts with the pattern, or ends with it
    :param pattern:
    :return:
    """
    if txt.endswith(pattern):
        return True
    if not txt.startswith(pattern):
        return False
    length = len(pattern)
    if len(txt) <= length:
        return True
    return txt[length] == "_" or txt[length].isupper()


def is_test_file(path: Path) -> bool:
    """
    Check if the given path points to a test/mock file
    :param path:
    :return:
    """
    return any(test_pattern in path.parts for test_pattern in TESTS_PATTERNS)


def is_test_contract(contract: "Contract") -> bool:
    """
    Check if the contract is a test/mock
    :param contract:
    :return:
    """
    return (
        _is_test_pattern(contract.name, "Test")
        or _is_test_pattern(contract.name, "Mock")
        or (
            contract.source_mapping.filename.absolute
            and is_test_file(Path(contract.source_mapping.filename.absolute))
        )
    )



from pathlib import Path

from slither import Slither
from slither.utils.arithmetic import unchecked_arithemtic_usage
from slither.slithir.operations import Binary, Unary, Assignment
from slither.slithir.variables.temporary import TemporaryVariable

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data" / "arithmetic_usage"


def test_arithmetic_usage(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.15")
    slither = Slither(Path(TEST_DATA_DIR, "test.sol").as_posix(), solc=solc_path)

    assert {
        f.source_mapping.content_hash for f in unchecked_arithemtic_usage(slither.contracts[0])
    } == {"2b4bc73cf59d486dd9043e840b5028b679354dd9", "e4ecd4d0fda7e762d29aceb8425f2c5d4d0bf962"}


def test_scope_is_checked(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.15")
    slither = Slither(Path(TEST_DATA_DIR, "unchecked_scope.sol").as_posix(), solc=solc_path)
    func = slither.get_contract_from_name("TestScope")[0].get_function_from_full_name(
        "scope(uint256)"
    )
    bin_op_is_checked = {}
    for node in func.nodes:
        for op in node.irs:
            if isinstance(op, (Binary, Unary)):
                bin_op_is_checked[op.lvalue] = op.node.scope.is_checked
            if isinstance(op, Assignment) and isinstance(op.rvalue, TemporaryVariable):
                if op.lvalue.name.startswith("checked"):
                    assert bin_op_is_checked[op.rvalue] is True
                else:
                    assert bin_op_is_checked[op.rvalue] is False



"""
Detect deletion on structure containing a mapping
"""

from slither.core.cfg.node import Node
from slither.core.declarations import Structure
from slither.core.declarations.contract import Contract
from slither.core.variables.variable import Variable
from slither.core.declarations.function_contract import FunctionContract
from slither.core.solidity_types import MappingType, UserDefinedType
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import Delete
from slither.utils.output import Output


class MappingDeletionDetection(AbstractDetector):
    """
    Mapping deletion detector
    """

    ARGUMENT = "mapping-deletion"
    HELP = "Deletion on mapping containing a structure"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#deletion-on-mapping-containing-a-structure"

    WIKI_TITLE = "Deletion on mapping containing a structure"
    WIKI_DESCRIPTION = "A deletion in a structure containing a mapping will not delete the mapping (see the [Solidity documentation](https://solidity.readthedocs.io/en/latest/types.html##delete)). The remaining data may be used to compromise the contract."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
    struct BalancesStruct{
        address owner;
        mapping(address => uint) balances;
    }
    mapping(address => BalancesStruct) public stackBalance;

    function remove() internal{
         delete stackBalance[msg.sender];
    }
```
`remove` deletes an item of `stackBalance`.
The mapping `balances` is never deleted, so `remove` does not work as intended."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = (
        "Use a lock mechanism instead of a deletion to disable structure containing a mapping."
    )

    @staticmethod
    def detect_mapping_deletion(
        contract: Contract,
    ) -> list[tuple[FunctionContract, Structure, Node]]:
        """Detect deletion on structure containing a mapping

        Returns:
            list (function, structure, node)
        """
        ret: list[tuple[FunctionContract, Structure, Node]] = []

        for f in contract.functions:
            for node in f.nodes:
                for ir in node.irs:
                    if isinstance(ir, Delete):
                        value = ir.variable
                        MappingDeletionDetection.check_if_mapping(value, ret, f, node)

        return ret

    @staticmethod
    def check_if_mapping(
        value: Variable,
        ret: list[tuple[FunctionContract, Structure, Node]],
        f: FunctionContract,
        node: Node,
    ):
        if isinstance(value.type, UserDefinedType) and isinstance(value.type.type, Structure):
            st = value.type.type
            if any(isinstance(e.type, MappingType) for e in st.elems.values()):
                ret.append((f, st, node))
                return
            for e in st.elems.values():
                MappingDeletionDetection.check_if_mapping(e, ret, f, node)

    def _detect(self) -> list[Output]:
        """Detect mapping deletion

        Returns:
            list: {'vuln', 'filename,'contract','func','struct''}
        """
        results = []
        for c in self.contracts:
            mapping = MappingDeletionDetection.detect_mapping_deletion(c)
            for func, struct, node in mapping:
                info: DETECTOR_INFO = [func, " deletes ", struct, " which contains a mapping:\n"]
                info += ["\t-", node, "\n"]

                res = self.generate_result(info)
                results.append(res)

        return results



"""
This module import all slither exceptions
"""

from slither.slithir.exceptions import SlithIRError as SlithIRError
from slither.solc_parsing.exceptions import ParsingError as ParsingError
from slither.solc_parsing.exceptions import VariableNotFound as VariableNotFound
from slither.core.exceptions import SlitherCoreError as SlitherCoreError
from slither.exceptions import SlitherException as SlitherException



from pathlib import Path
import shutil
import re

import pytest

from slither import Slither
from slither.exceptions import SlitherException


TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"

foundry_available = shutil.which("forge") is not None
project_ready = Path(TEST_DATA_DIR, "test_change/lib/forge-std").exists()


@pytest.mark.skipif(
    not foundry_available or not project_ready, reason="requires Foundry and project setup"
)
def test_diagnostic():
    test_file_directory = TEST_DATA_DIR / "test_change"

    sl = Slither(test_file_directory.as_posix())
    assert len(sl.compilation_units) == 1

    counter_file = test_file_directory / "src" / "Counter.sol"
    shutil.copy(counter_file, counter_file.with_suffix(".bak"))

    with counter_file.open("r") as file:
        content = file.read()

    with counter_file.open("w") as file:
        file.write(re.sub(r"//START.*?//END\n?", "", content, flags=re.DOTALL))

    with pytest.raises(SlitherException):
        Slither(test_file_directory.as_posix(), ignore_compile=True)

    # Restore the original counter so the test is idempotent
    Path(counter_file.with_suffix(".bak")).rename(counter_file)



from collections.abc import Callable
from dataclasses import dataclass

from slither.tools.doctor.checks.paths import check_slither_path
from slither.tools.doctor.checks.platform import compile_project, detect_platform
from slither.tools.doctor.checks.versions import show_versions


@dataclass(slots=True)
class Check:
    title: str
    function: Callable[..., None]


ALL_CHECKS: list[Check] = [
    Check("PATH configuration", check_slither_path),
    Check("Software versions", show_versions),
    Check("Project platform", detect_platform),
    Check("Project compilation", compile_project),
]



from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output
from .arbitrary_send_erc20 import ArbitrarySendErc20


class ArbitrarySendErc20Permit(AbstractDetector):
    """
    Detect when `msg.sender` is not used as `from` in transferFrom along with the use of permit.
    """

    ARGUMENT = "arbitrary-send-erc20-permit"
    HELP = "transferFrom uses arbitrary from with permit"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#arbitrary-from-in-transferfrom-used-with-permit"

    WIKI_TITLE = "Arbitrary `from` in transferFrom used with permit"
    WIKI_DESCRIPTION = (
        "Detect when `msg.sender` is not used as `from` in transferFrom and permit is used."
    )
    WIKI_EXPLOIT_SCENARIO = """
```solidity
    function bad(address from, uint256 value, uint256 deadline, uint8 v, bytes32 r, bytes32 s, address to) public {
        erc20.permit(from, address(this), value, deadline, v, r, s);
        erc20.transferFrom(from, to, value);
    }
```
If an ERC20 token does not implement permit and has a fallback function e.g. WETH, transferFrom allows an attacker to transfer all tokens approved for this contract."""

    WIKI_RECOMMENDATION = """
Ensure that the underlying ERC20 token correctly implements a permit function.
"""

    def _detect(self) -> list[Output]:
        """"""
        results: list[Output] = []

        arbitrary_sends = ArbitrarySendErc20(self.compilation_unit)
        arbitrary_sends.detect()
        for node in arbitrary_sends.permit_results:
            func = node.function
            info: DETECTOR_INFO = [
                func,
                " uses arbitrary from in transferFrom in combination with permit: ",
                node,
                "\n",
            ]
            res = self.generate_result(info)
            results.append(res)

        return results



#!/usr/bin/env python3

import argparse
import logging
import sys

from crytic_compile import cryticparser

from slither.tools.similarity.info import info
from slither.tools.similarity.test import test
from slither.tools.similarity.train import train
from slither.tools.similarity.plot import plot

logging.basicConfig()
logger = logging.getLogger("Slither-simil")

modes = ["info", "test", "train", "plot"]


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Code similarity detection tool. For usage, see https://github.com/crytic/slither/wiki/Code-Similarity-detector"
    )

    parser.add_argument("mode", help="|".join(modes))

    parser.add_argument("model", help="model.bin")

    parser.add_argument("--filename", action="store", dest="filename", help="contract.sol")

    parser.add_argument("--fname", action="store", dest="fname", help="Target function")

    parser.add_argument("--ext", action="store", dest="ext", help="Extension to filter contracts")

    parser.add_argument(
        "--nsamples",
        action="store",
        type=int,
        dest="nsamples",
        help="Number of contract samples used for training",
    )

    parser.add_argument(
        "--ntop",
        action="store",
        type=int,
        dest="ntop",
        default=10,
        help="Number of more similar contracts to show for testing",
    )

    parser.add_argument(
        "--input", action="store", dest="input", help="File or directory used as input"
    )

    parser.add_argument(
        "--version",
        help="displays the current version",
        version="0.0",
        action="version",
    )

    cryticparser.init(parser)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    return args


# endregion
###################################################################################
###################################################################################
# region Main
###################################################################################
###################################################################################


def main() -> None:
    args = parse_args()

    default_log = logging.INFO
    logger.setLevel(default_log)

    mode = args.mode

    if mode == "info":
        info(args)
    elif mode == "train":
        train(args)
    elif mode == "test":
        test(args)
    elif mode == "plot":
        plot(args)
    else:
        to_log = f"Invalid mode!. It should be one of these: {', '.join(modes)}"
        logger.error(to_log)
        sys.exit(-1)


if __name__ == "__main__":
    main()

# endregion



from slither.core.solidity_types.elementary_type import (
    ElementaryType,
    ElementaryTypeName,
)  # TODO rename solidity type
from slither.core.solidity_types.array_type import ArrayType
from slither.core.solidity_types.mapping_type import MappingType
from slither.core.solidity_types.user_defined_type import UserDefinedType
from slither.core.declarations import FunctionContract, Contract
from slither.vyper_parsing.ast.types import Name, Subscript, Call, Index, Tuple
from slither.solc_parsing.exceptions import ParsingError


def parse_type(
    annotation: Name | Subscript | Call | Tuple,
    caller_context: FunctionContract | Contract,
):
    from slither.vyper_parsing.expressions.expression_parsing import parse_expression

    if isinstance(caller_context, FunctionContract):
        contract = caller_context.contract
    else:
        contract = caller_context

    assert isinstance(annotation, (Name, Subscript, Call, Tuple))

    if isinstance(annotation, Name):
        name = annotation.id
        lname = name.lower()  # map `String` to string
        if lname in ElementaryTypeName:
            return ElementaryType(lname)

        if name in contract.structures_as_dict:
            return UserDefinedType(contract.structures_as_dict[name])

        if name in contract.enums_as_dict:
            return UserDefinedType(contract.enums_as_dict[name])

        if name in contract.file_scope.contracts:
            return UserDefinedType(contract.file_scope.contracts[name])

        if name in contract.file_scope.structures:
            return UserDefinedType(contract.file_scope.structures[name])
    elif isinstance(annotation, Subscript):
        assert isinstance(annotation.slice, Index)
        # This is also a strange construct... https://github.com/vyperlang/vyper/issues/3577
        if isinstance(annotation.slice.value, Tuple):
            assert isinstance(annotation.value, Name)
            if annotation.value.id == "DynArray":
                type_ = parse_type(annotation.slice.value.elements[0], caller_context)
                length = parse_expression(annotation.slice.value.elements[1], caller_context)
                return ArrayType(type_, length)
            if annotation.value.id == "HashMap":
                type_from = parse_type(annotation.slice.value.elements[0], caller_context)
                type_to = parse_type(annotation.slice.value.elements[1], caller_context)

                return MappingType(type_from, type_to)

        elif isinstance(annotation.value, Subscript):
            type_ = parse_type(annotation.value, caller_context)

        elif isinstance(annotation.value, Name):
            # TODO it is weird that the ast_type is `Index` when it's a type annotation and not an expression, so we grab the value. https://github.com/vyperlang/vyper/issues/3577
            type_ = parse_type(annotation.value, caller_context)
            if annotation.value.id == "String":
                # This is an elementary type
                return type_

        length = parse_expression(annotation.slice.value, caller_context)
        return ArrayType(type_, length)

    elif isinstance(annotation, Call):
        # TODO event variable represented as Call https://github.com/vyperlang/vyper/issues/3579
        return parse_type(annotation.args[0], caller_context)

    elif isinstance(annotation, Tuple):
        # Vyper has tuple types like python x = f() where f() -> (y,z)
        # and tuple elements can be unpacked like x[0]: y and x[1]: z.
        # We model these as a struct and unpack each index into a field
        # e.g. accessing the 0th element is translated as x._0
        from slither.core.declarations.structure import Structure
        from slither.core.variables.structure_variable import StructureVariable

        st = Structure(caller_context.compilation_unit)
        st.set_offset("-1:-1:-1", caller_context.compilation_unit)
        st.name = "FAKE_TUPLE"
        for idx, elem_info in enumerate(annotation.elements):
            elem = StructureVariable()
            elem.type = parse_type(elem_info, caller_context)
            elem.name = f"_{idx}"
            elem.set_structure(st)
            elem.set_offset("-1:-1:-1", caller_context.compilation_unit)
            st.elems[elem.name] = elem
            st.add_elem_in_order(elem.name)
            st.name += elem.name

        return UserDefinedType(st)

    raise ParsingError(f"Type name not found {name} context {caller_context}")



import argparse
import logging
from collections import defaultdict
from typing import Any
from collections.abc import Callable

from crytic_compile import cryticparser

from slither import Slither
from slither.core.declarations import Contract
from slither.utils.erc import ERCS
from slither.utils.output import output_to_json
from .erc.erc1155 import check_erc1155
from .erc.erc20 import check_erc20
from .erc.ercs import generic_erc_checks

logging.basicConfig()
logging.getLogger("Slither").setLevel(logging.INFO)

logger = logging.getLogger("Slither-conformance")
logger.setLevel(logging.INFO)

ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(message)s")
logger.addHandler(ch)
logger.handlers[0].setFormatter(formatter)
logger.propagate = False

ADDITIONAL_CHECKS: dict[str, Callable[[Contract, dict[str, list]], dict[str, list]]] = {
    "ERC20": check_erc20,
    "ERC1155": check_erc1155,
}


def parse_args() -> argparse.Namespace:
    """
    Parse the underlying arguments for the program.
    :return: Returns the arguments for the program.
    """
    parser = argparse.ArgumentParser(
        description="Check the ERC 20 conformance",
        usage="slither-check-erc project contractName",
    )

    parser.add_argument("project", help="The codebase to be tested.")

    parser.add_argument(
        "contract_name",
        help="The name of the contract. Specify the first case contract that follow the standard. Derived contracts will be checked.",
    )

    parser.add_argument(
        "--erc",
        help=f"ERC to be tested, available {','.join(ERCS.keys())} (default ERC20)",
        action="store",
        default="erc20",
    )

    parser.add_argument(
        "--json",
        help='Export the results as a JSON file ("--json -" to export to stdout)',
        action="store",
        default=False,
    )

    # Add default arguments from crytic-compile
    cryticparser.init(parser)

    return parser.parse_args()


def _log_error(err: Any, args: argparse.Namespace) -> None:
    if args.json:
        output_to_json(args.json, str(err), {"erc-conformance": []})

    logger.error(err)


def main() -> None:
    args = parse_args()

    # Perform slither analysis on the given filename
    slither = Slither(args.project, **vars(args))

    ret: dict[str, list] = defaultdict(list)

    if args.erc.upper() in ERCS:
        contracts = slither.get_contract_from_name(args.contract_name)

        if len(contracts) != 1:
            err = f"Contract not found: {args.contract_name}"
            _log_error(err, args)
            return
        contract = contracts[0]
        # First elem is the function, second is the event
        erc = ERCS[args.erc.upper()]
        generic_erc_checks(contract, erc[0], erc[1], ret)

        if args.erc.upper() in ADDITIONAL_CHECKS:
            ADDITIONAL_CHECKS[args.erc.upper()](contract, ret)

    else:
        err = f"Incorrect ERC selected {args.erc}"
        _log_error(err, args)
        return

    if args.json:
        output_to_json(args.json, None, {"erc-conformance": ret})


if __name__ == "__main__":
    main()



from enum import Enum
from typing import NamedTuple


class PropertyType(Enum):
    CODE_QUALITY = 1
    LOW_SEVERITY = 2
    MEDIUM_SEVERITY = 3
    HIGH_SEVERITY = 4


class PropertyReturn(Enum):
    SUCCESS = 1
    FAIL_OR_THROW = 2
    FAIL = 3
    THROW = 4


class PropertyCaller(Enum):
    OWNER = 1
    SENDER = 2
    ATTACKER = 3
    ALL = 4  # If all the actors should call the function. Typically if the test uses msg.sender
    ANY = 5  # If the caller does not matter


class Property(NamedTuple):
    name: str
    content: str
    type: PropertyType
    return_type: PropertyReturn
    is_unit_test: bool
    caller: PropertyCaller  # Only for unit tests. Echidna will try all the callers
    is_property_test: bool
    description: str


def property_to_solidity(p: Property):
    return f"\tfunction {p.name} public returns(bool){{{p.content}\n\t}}\n"



from typing import TYPE_CHECKING, Any

from slither.core.expressions.expression import Expression
from slither.core.solidity_types.elementary_type import Fixed, Int, Ufixed, Uint
from slither.utils.arithmetic import convert_subdenomination
from slither.utils.integer_conversion import convert_string_to_int

if TYPE_CHECKING:
    from slither.core.solidity_types.type import Type


class Literal(Expression):
    def __init__(
        self, value: int | str, custom_type: "Type", subdenomination: str | None = None
    ) -> None:
        super().__init__()
        self._value = value
        self._type = custom_type
        self._subdenomination = subdenomination

    @property
    def value(self) -> int | str:
        return self._value

    @property
    def converted_value(self) -> int | str:
        """Return the value of the literal, accounting for subdenomination e.g. ether"""
        if self.subdenomination:
            return convert_subdenomination(self._value, self.subdenomination)
        return self._value

    @property
    def type(self) -> "Type":
        return self._type

    @property
    def subdenomination(self) -> str | None:
        return self._subdenomination

    def __str__(self) -> str:
        if self.subdenomination:
            return str(self.converted_value)

        if self.type in Int + Uint + Fixed + Ufixed + ["address"]:
            return str(convert_string_to_int(self._value))

        # be sure to handle any character
        return str(self._value)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Literal):
            return False
        return (self.value, self.subdenomination) == (other.value, other.subdenomination)



import logging
import os
from argparse import ArgumentParser
from pathlib import Path

from slither.utils.command_line import defaults_flag_in_config

logger = logging.getLogger("Slither")


def init_parser(parser: ArgumentParser, always_enable_codex: bool = False) -> None:
    """
    Init the cli arg with codex features

    Args:
        parser:
        always_enable_codex (Optional(bool)): if true, --codex is not enabled

    Returns:

    """
    group_codex = parser.add_argument_group("Codex (https://beta.openai.com/docs/guides/code)")

    if not always_enable_codex:
        group_codex.add_argument(
            "--codex",
            help="Enable codex (require an OpenAI API Key)",
            action="store_true",
            default=defaults_flag_in_config["codex"],
        )

    group_codex.add_argument(
        "--codex-log",
        help="Log codex queries (in crytic_export/codex/)",
        action="store_true",
        default=False,
    )

    group_codex.add_argument(
        "--codex-contracts",
        help="Comma separated list of contracts to submit to OpenAI Codex",
        action="store",
        default=defaults_flag_in_config["codex_contracts"],
    )

    group_codex.add_argument(
        "--codex-model",
        help="Name of the Codex model to use (affects pricing).  Defaults to 'text-davinci-003'",
        action="store",
        default=defaults_flag_in_config["codex_model"],
    )

    group_codex.add_argument(
        "--codex-temperature",
        help="Temperature to use with Codex.  Lower number indicates a more precise answer while higher numbers return more creative answers.  Defaults to 0",
        action="store",
        default=defaults_flag_in_config["codex_temperature"],
    )

    group_codex.add_argument(
        "--codex-max-tokens",
        help="Maximum amount of tokens to use on the response.  This number plus the size of the prompt can be no larger than the limit (4097 for text-davinci-003)",
        action="store",
        default=defaults_flag_in_config["codex_max_tokens"],
    )

    group_codex.add_argument(
        "--codex-organization",
        help="Codex organization",
        action="store",
        default=None,
    )


# TODO: investigate how to set the correct return type
# So that the other modules can work with openai
def openai_module():  # type: ignore
    """
    Return the openai module
    Consider checking the usage of open (slither.codex_enabled) before using this function

    Returns:
        Optional[the openai module]
    """
    try:
        import openai

        api_key = os.getenv("OPENAI_API_KEY")
        if api_key is None:
            logger.info(
                "Please provide an Open API Key in OPENAI_API_KEY (https://beta.openai.com/account/api-keys)"
            )
            return None
        openai.api_key = api_key
    except ImportError:
        logger.info("OpenAI was not installed")  # type: ignore
        logger.info('run "pip install openai"')
        return None
    return openai


def log_codex(filename: str, prompt: str) -> None:
    """
    Log the prompt in crytic/export/codex/filename
    Append to the file

    Args:
        filename: filename to write to
        prompt: prompt to write

    Returns:
        None
    """

    Path("crytic_export/codex").mkdir(parents=True, exist_ok=True)

    with open(Path("crytic_export/codex", filename), "a", encoding="utf8") as file:
        file.write(prompt)
        file.write("\n")



"""
Module detecting unimplemented interfaces

Collect all the interfaces
Check for contracts which implement all interface functions but do not explicitly derive from those interfaces.
"""

from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.core.declarations.contract import Contract
from slither.utils.output import Output


class MissingInheritance(AbstractDetector):
    """
    Unimplemented interface detector
    """

    ARGUMENT = "missing-inheritance"
    HELP = "Missing inheritance"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#missing-inheritance"
    WIKI_TITLE = "Missing inheritance"
    WIKI_DESCRIPTION = "Detect missing inheritance."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
interface ISomething {
    function f1() external returns(uint);
}

contract Something {
    function f1() external returns(uint){
        return 42;
    }
}
```
`Something` should inherit from `ISomething`.
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Inherit from the missing interface or contract."

    @staticmethod
    def detect_unimplemented_interface(
        contract: Contract, interfaces: list[Contract]
    ) -> list[Contract]:
        """
        Detects if contract intends to implement one of the interfaces but does not explicitly do so by deriving from it
        :param contract: The contract to check
        :param interfaces: List of all the interfaces
        :return: Interfaces likely intended to implement by the contract
        """

        intended_interfaces: list[Contract] = []
        sigs_contract = {f.full_name for f in contract.functions_entry_points}

        if not sigs_contract:
            return intended_interfaces

        for interface in interfaces:
            # If contract already inherits from interface, skip that interface
            if interface in contract.inheritance:
                continue

            sigs_interface = {f.full_name for f in interface.functions_entry_points}

            # Contract should implement all the functions of the intended interface
            if not sigs_interface.issubset(sigs_contract):
                continue

            # A parent contract inherited should not implement a superset of intended interface
            # This is because in the following:
            # interface ERC20_interface:
            #    - balanceOf(uint) -> uint
            #    - transfer(address, address) -> bool
            # contract ForeignToken:
            #    - balanceOf(uint) -> uint
            # contract MyERC20Implementation is ERC20_interface
            #
            # We do not want MyERC20Implementation to be declared as missing ForeignToken interface
            intended_interface_is_subset_parent = False
            for parent in contract.inheritance:
                sigs_parent = {f.full_name for f in parent.functions_entry_points}
                if sigs_interface.issubset(sigs_parent):
                    intended_interface_is_subset_parent = True
                    break

            if not intended_interface_is_subset_parent:
                # Should not be a subset of an earlier determined intended_interface or derive from it
                intended_interface_is_subset_intended = False
                for intended_interface in list(intended_interfaces):
                    sigs_intended_interface = {
                        f.full_name for f in intended_interface.functions_entry_points
                    }
                    if (
                        sigs_interface.issubset(sigs_intended_interface)
                        or interface in intended_interface.inheritance
                    ):
                        intended_interface_is_subset_intended = True
                        break

                    # If superset of an earlier determined intended_interface or derives from it,
                    # remove the intended_interface
                    if (
                        sigs_intended_interface.issubset(sigs_interface)
                        or intended_interface in interface.inheritance
                    ):
                        intended_interfaces.remove(intended_interface)

                if not intended_interface_is_subset_intended:
                    intended_interfaces.append(interface)

        return intended_interfaces

    def _detect(self) -> list[Output]:
        """Detect unimplemented interfaces
        Returns:
            list: {'contract'}
        """

        # Collect all the interfaces
        # Here interface can be "interface" from solidity, or contracts with only functions declaration
        # Skip interfaces without functions
        interfaces = [
            contract
            for contract in self.compilation_unit.contracts
            if contract.is_signature_only()
            and any(not f.is_constructor_variables for f in contract.functions)
        ]

        # Check derived contracts for missing interface implementations
        results = []
        for contract in self.compilation_unit.contracts_derived:
            # Skip interfaces
            if contract in interfaces:
                continue
            intended_interfaces = self.detect_unimplemented_interface(contract, interfaces)
            for interface in intended_interfaces:
                info: DETECTOR_INFO = [contract, " should inherit from ", interface, "\n"]
                res = self.generate_result(info)
                results.append(res)
        return results



from typing import TYPE_CHECKING

from slither.core.declarations.contract_level import ContractLevel
from slither.core.declarations import Enum

if TYPE_CHECKING:
    from slither.core.declarations import Contract


class EnumContract(Enum, ContractLevel):
    def is_declared_by(self, contract: "Contract") -> bool:
        """
        Check if the element is declared by the contract
        :param contract:
        :return:
        """
        return self.contract == contract



from slither import Slither

slither = Slither("coin.sol")

whitelist = ["balanceOf(address)"]

for contract in slither.contracts:
    for function in contract.functions:
        if function.full_name in whitelist:
            continue
        if function.is_constructor:
            continue
        if function.visibility in ["public", "external"]:
            if not "onlyOwner()" in [m.full_name for m in function.modifiers]:
                print(f"{function.full_name} is unprotected!")



from typing import Any

from slither.core.variables import Variable
from slither.slithir.operations.operation import Operation


class OperationWithLValue(Operation):
    """
    Operation with a lvalue
    """

    def __init__(self) -> None:
        super().__init__()

        self._lvalue: Variable | None = None

    @property
    def lvalue(self) -> Variable | None:
        return self._lvalue

    @lvalue.setter
    def lvalue(self, lvalue: Variable) -> None:
        self._lvalue = lvalue

    @property
    def used(self) -> list[Any | None]:
        return self.read + [self.lvalue]



from pathlib import Path


from slither import Slither
from slither.core.variables.state_variable import StateVariable

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"
CONTRACT_DECL_TEST_ROOT = Path(TEST_DATA_DIR, "contract_declaration")


def test_abstract_contract(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.0")
    slither = Slither(Path(CONTRACT_DECL_TEST_ROOT, "abstract.sol").as_posix(), solc=solc_path)
    explicit_abstract = slither.contracts[0]
    assert not explicit_abstract.is_fully_implemented
    assert explicit_abstract.is_abstract

    solc_path = solc_binary_path("0.5.0")
    slither = Slither(
        Path(CONTRACT_DECL_TEST_ROOT, "implicit_abstract.sol").as_posix(), solc=solc_path
    )
    implicit_abstract = slither.get_contract_from_name("ImplicitAbstract")[0]
    assert not implicit_abstract.is_fully_implemented
    # This only is expected to work for newer versions of Solidity
    assert not implicit_abstract.is_abstract

    slither = Slither(
        Path(CONTRACT_DECL_TEST_ROOT, "implicit_abstract.sol").as_posix(),
        solc_force_legacy_json=True,
        solc=solc_path,
    )
    implicit_abstract = slither.get_contract_from_name("ImplicitAbstract")[0]
    assert not implicit_abstract.is_fully_implemented
    # This only is expected to work for newer versions of Solidity
    assert not implicit_abstract.is_abstract


def test_concrete_contract(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.0")
    slither = Slither(Path(CONTRACT_DECL_TEST_ROOT, "concrete.sol").as_posix(), solc=solc_path)
    concrete = slither.get_contract_from_name("Concrete")[0]
    assert concrete.is_fully_implemented
    assert not concrete.is_abstract

    solc_path = solc_binary_path("0.5.0")
    slither = Slither(
        Path(CONTRACT_DECL_TEST_ROOT, "concrete_old.sol").as_posix(),
        solc_force_legacy_json=True,
        solc=solc_path,
    )
    concrete_old = slither.get_contract_from_name("ConcreteOld")[0]
    assert concrete_old.is_fully_implemented
    assert not concrete_old.is_abstract


def test_private_variable(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.15")
    slither = Slither(
        Path(CONTRACT_DECL_TEST_ROOT, "private_variable.sol").as_posix(), solc=solc_path
    )
    contract_c = slither.get_contract_from_name("C")[0]
    f = contract_c.functions[0]
    var_read = f.variables_read[0]
    assert isinstance(var_read, StateVariable)
    assert str(var_read.contract) == "B"



"""
Module printing all the state-changing entry point functions and their variables of the contracts
"""

from pathlib import Path
from slither.printers.abstract_printer import AbstractPrinter
from slither.core.declarations.function_contract import FunctionContract
from slither.utils.colors import Colors
from slither.utils.output import Output
from slither.utils.myprettytable import MyPrettyTable
from slither.utils.tests_pattern import is_test_file


class PrinterEntryPoints(AbstractPrinter):
    ARGUMENT = "entry-points"
    HELP = "Print all the state-changing entry point functions and their variables of the contracts"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#entry-points"

    def output(self, _filename) -> Output:
        """
        _filename is not used
        Args:
            _filename(string)
        """
        all_contracts = []

        for contract in sorted(
            (
                c
                for c in self.slither.contracts_derived
                if not c.is_test
                and not c.is_from_dependency()
                and not is_test_file(Path(c.source_mapping.filename.absolute))
                and not c.is_interface
                and not c.is_library
                and not c.is_abstract
            ),
            key=lambda x: x.name,
        ):
            entry_points = [
                f
                for f in contract.functions
                if (
                    f.visibility in ["public", "external"]
                    and isinstance(f, FunctionContract)
                    and not f.view
                    and not f.pure
                    and not f.is_shadowed
                )
            ]

            if not entry_points:
                continue

            inheritance_chain = self._get_inheritance_chain(contract)
            inheritance_str = f" is {' is '.join(inheritance_chain)}" if inheritance_chain else ""

            contract_info = [
                f"\n\nContract {Colors.BOLD}{Colors.YELLOW}{contract.name}{Colors.END}{inheritance_str}"
                f" ({contract.source_mapping})"
            ]

            variables_info = self._get_variables_info(contract)
            if variables_info:
                var_table = MyPrettyTable(["Variables", "Types", "Inherited From"])
                for var_name, var_type, inherited_from in variables_info:
                    var_table.add_row(
                        [
                            f"{Colors.BOLD}{Colors.BLUE}{var_name}{Colors.END}",
                            f"{Colors.GREEN}{var_type}{Colors.END}",
                            (
                                f"{Colors.MAGENTA}{inherited_from}{Colors.END}"
                                if inherited_from
                                else ""
                            ),
                        ]
                    )
                contract_info.append(str(var_table))

            func_table = MyPrettyTable(["Functions", "Modifiers", "Inherited From"])

            self._add_function_rows(func_table, entry_points, contract)

            contract_info.append(str(func_table))
            all_contracts.append("\n".join(contract_info))

        info = "\n".join(all_contracts) if all_contracts else ""
        self.info(info)
        return self.generate_output(info)

    def _get_inheritance_chain(self, contract):
        """Get the full inheritance chain for a contract"""
        inheritance_chain = []
        for base in contract.inheritance:
            if not base.is_interface and not base.is_library:
                inheritance_chain.append(base.name)
        return inheritance_chain

    def _get_variables_info(self, contract):
        """Get all state variables with their types and inheritance info"""
        variables_info = []

        for variable in contract.storage_variables_ordered:
            var_type = str(variable.type)
            inherited_from = variable.contract.name if variable.contract != contract else ""
            variables_info.append((variable.name, var_type, inherited_from))

        return variables_info

    def _add_function_rows(self, func_table, entry_points, contract):
        """Add function rows to the functions table"""
        for f in sorted(
            entry_points,
            key=lambda x, contract=contract: (
                x.contract_declarer != contract,
                x.contract_declarer.name if x.contract_declarer != contract else "",
                x.source_mapping.start,
            ),
        ):
            name_parts = f.full_name.split("(", 1)
            inherited = f.contract_declarer.name if f.contract_declarer != contract else ""
            modifiers = ", ".join(
                [m.name for m in f.modifiers] + (["payable"] if f.payable else [])
            )

            if f.is_constructor or f.name in ["receive", "fallback"]:
                function_color = f"{Colors.BOLD}{Colors.MAGENTA}"
            else:
                function_color = f"{Colors.BOLD}{Colors.RED}"

            function_name = name_parts[0]

            func_table.add_row(
                [
                    f"{function_color}{function_name}{Colors.END}({name_parts[1]}",
                    f"{Colors.GREEN}{modifiers}{Colors.END}" if modifiers else "",
                    f"{Colors.MAGENTA}{inherited}{Colors.END}" if inherited else "",
                ]
            )



from typing import Any

from slither.core.declarations import Function
from slither.core.declarations.contract import Contract
from slither.core.variables import Variable
from slither.core.solidity_types import UserDefinedType
from slither.slithir.operations import Call, OperationWithLValue
from slither.slithir.utils.utils import is_valid_lvalue
from slither.slithir.variables.temporary import TemporaryVariable
from slither.slithir.variables.temporary_ssa import TemporaryVariableSSA


class NewContract(Call, OperationWithLValue):
    def __init__(
        self,
        contract_name: UserDefinedType,
        lvalue: TemporaryVariableSSA | TemporaryVariable,
        names: list[str] | None = None,
    ) -> None:
        """
        #### Parameters
        names -
            For calls of the form f({argName1 : arg1, ...}), the names of parameters listed in call order.
            Otherwise, None.
        """
        assert isinstance(contract_name.type, Contract), (
            f"contract_name is {contract_name} of type {type(contract_name)}"
        )
        assert is_valid_lvalue(lvalue)
        super().__init__(names=names)
        self._contract_name = contract_name
        # todo create analyze to add the contract instance
        self._lvalue = lvalue
        self._callid = None  # only used if gas/value != 0
        self._call_value = None
        self._call_salt = None

    @property
    def call_value(self):
        return self._call_value

    @call_value.setter
    def call_value(self, v):
        self._call_value = v

    @property
    def call_id(self):
        return self._callid

    @call_id.setter
    def call_id(self, c):
        self._callid = c

    @property
    def call_salt(self):
        return self._call_salt

    @call_salt.setter
    def call_salt(self, s):
        self._call_salt = s

    @property
    def contract_name(self) -> UserDefinedType:
        return self._contract_name

    @property
    def read(self) -> list[Any]:
        all_read = [self.call_salt, self.call_value] + self._unroll(self.arguments)
        # remove None
        return [x for x in all_read if x]

    @property
    def contract_created(self) -> Contract:
        return self.contract_name.type

    ###################################################################################
    ###################################################################################
    # region Analyses
    ###################################################################################
    ###################################################################################

    def can_reenter(self, callstack: list[Function | Variable] | None = None) -> bool:
        """
        Must be called after slithIR analysis pass
        For Solidity > 0.5, filter access to public variables and constant/pure/view
        For call to this. check if the destination can re-enter
        :param callstack: check for recursion
        :return: bool
        """
        callstack = [] if callstack is None else callstack
        constructor = self.contract_created.constructor
        if constructor is None:
            return False
        if constructor in callstack:
            return False
        callstack = callstack + [constructor]
        return constructor.can_reenter(callstack)

    def can_send_eth(self) -> bool:
        """
        Must be called after slithIR analysis pass
        :return: bool
        """
        return self._call_value is not None

    # endregion

    def __str__(self) -> str:
        options = ""
        if self.call_value:
            options = f"value:{self.call_value} "
        if self.call_salt:
            options += f"salt:{self.call_salt} "
        args = [str(a) for a in self.arguments]
        lvalue = self.lvalue
        return f"{lvalue}({lvalue.type}) = new {self.contract_name}({','.join(args)}) {options}"



from typing import TYPE_CHECKING

from slither.core.variables.top_level_variable import TopLevelVariable
from slither.solc_parsing.variables.variable_declaration import VariableDeclarationSolc
from slither.solc_parsing.declarations.caller_context import CallerContextExpression

if TYPE_CHECKING:
    from slither.solc_parsing.slither_compilation_unit_solc import SlitherCompilationUnitSolc
    from slither.core.compilation_unit import SlitherCompilationUnit


class TopLevelVariableSolc(VariableDeclarationSolc, CallerContextExpression):
    def __init__(
        self,
        variable: TopLevelVariable,
        variable_data: dict,
        slither_parser: "SlitherCompilationUnitSolc",
    ) -> None:
        super().__init__(variable, variable_data)
        self._slither_parser = slither_parser

    @property
    def is_compact_ast(self) -> bool:
        return self._slither_parser.is_compact_ast

    @property
    def compilation_unit(self) -> "SlitherCompilationUnit":
        return self._slither_parser.compilation_unit

    def get_key(self) -> str:
        return self._slither_parser.get_key()

    @property
    def slither_parser(self) -> "SlitherCompilationUnitSolc":
        return self._slither_parser

    @property
    def underlying_variable(self) -> TopLevelVariable:
        # Todo: Not sure how to overcome this with mypy
        assert isinstance(self._variable, TopLevelVariable)
        return self._variable



from slither.core.solidity_types import ElementaryType
from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.utils.utils import is_valid_lvalue, is_valid_rvalue
from slither.core.variables.local_variable import LocalVariable
from slither.core.variables.state_variable import StateVariable
from slither.slithir.variables.local_variable import LocalIRVariable
from slither.slithir.variables.reference import ReferenceVariable
from slither.slithir.variables.reference_ssa import ReferenceVariableSSA
from slither.slithir.variables.state_variable import StateIRVariable


class Length(OperationWithLValue):
    def __init__(
        self,
        value: StateVariable | LocalIRVariable | LocalVariable | StateIRVariable,
        lvalue: ReferenceVariable | ReferenceVariableSSA,
    ) -> None:
        super().__init__()
        assert is_valid_rvalue(value)
        assert is_valid_lvalue(lvalue)
        self._value = value
        self._lvalue = lvalue
        lvalue.set_type(ElementaryType("uint256"))

    @property
    def read(self) -> list[LocalVariable | StateVariable | LocalIRVariable | StateIRVariable]:
        return [self._value]

    @property
    def value(self) -> StateVariable | LocalVariable:
        return self._value

    def __str__(self):
        return f"{self.lvalue} -> LENGTH {self.value}"



from .contract import Contract
from .enum import Enum
from .event import Event
from .event_contract import EventContract
from .event_top_level import EventTopLevel
from .function import Function
from .import_directive import Import
from .modifier import Modifier
from .pragma_directive import Pragma
from .solidity_variables import (
    SolidityVariable,
    SolidityVariableComposed,
    SolidityFunction,
)
from .structure import Structure
from .enum_contract import EnumContract
from .enum_top_level import EnumTopLevel
from .structure_contract import StructureContract
from .structure_top_level import StructureTopLevel
from .function_contract import FunctionContract
from .function_top_level import FunctionTopLevel
from .custom_error_contract import CustomErrorContract
from .custom_error_top_level import CustomErrorTopLevel
from .custom_error import CustomError
from .solidity_import_placeholder import SolidityImportPlaceHolder



"""
Module detecting unused return values from external calls
"""

from slither.core.cfg.node import Node, NodeType
from slither.core.declarations import Function
from slither.core.declarations.function_contract import FunctionContract
from slither.core.variables.state_variable import StateVariable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import HighLevelCall, Assignment, Unpack, Operation
from slither.slithir.variables import TupleVariable
from slither.utils.output import Output


class UnusedReturnValues(AbstractDetector):
    """
    If the return value of a function is never used, it's likely to be bug
    """

    ARGUMENT = "unused-return"
    HELP = "Unused return values"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#unused-return"

    WIKI_TITLE = "Unused return"
    WIKI_DESCRIPTION = (
        "The return value of an external call is not stored in a local or state variable."
    )

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract MyConc{
    using SafeMath for uint;
    function my_func(uint a, uint b) public{
        a.add(b);
    }
}
```
`MyConc` calls `add` of `SafeMath`, but does not store the result in `a`. As a result, the computation has no effect."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Ensure that all the return values of the function calls are used."

    def _is_instance(self, ir: Operation) -> bool:
        return (
            isinstance(ir, HighLevelCall)
            and (
                (
                    isinstance(ir.function, Function)
                    and ir.function.solidity_signature
                    not in ["transfer(address,uint256)", "transferFrom(address,address,uint256)"]
                )
                or not isinstance(ir.function, Function)
            )
        ) or (ir.node.type == NodeType.TRY and isinstance(ir, (Assignment, Unpack)))

    def detect_unused_return_values(self, f: FunctionContract) -> list[Node]:
        """
            Return the nodes where the return value of a call is unused
        Args:
            f (Function)
        Returns:
            list(Node)
        """
        values_returned = []
        nodes_origin = {}

        for n in f.nodes:
            for ir in n.irs:
                if self._is_instance(ir):
                    # if a return value is stored in a state variable, it's ok
                    if ir.lvalue and not isinstance(ir.lvalue, StateVariable):
                        values_returned.append((ir.lvalue, None))
                        nodes_origin[ir.lvalue] = ir
                        if isinstance(ir.lvalue, TupleVariable):
                            # we iterate the number of elements the tuple has
                            # and add a (variable, index) in values_returned for each of them
                            for index in range(len(ir.lvalue.type)):
                                values_returned.append((ir.lvalue, index))
                for read in ir.read:
                    remove = (read, ir.index) if isinstance(ir, Unpack) else (read, None)
                    if remove in values_returned:
                        # this is needed to remove the tuple variable when the first time one of its element is used
                        if remove[1] is not None and (remove[0], None) in values_returned:
                            values_returned.remove((remove[0], None))
                        values_returned.remove(remove)
        return [nodes_origin[value].node for (value, _) in values_returned]

    def _detect(self) -> list[Output]:
        """Detect high level calls which return a value that are never used"""
        results = []
        for c in self.compilation_unit.contracts_derived:
            for f in c.functions_and_modifiers:
                unused_return = self.detect_unused_return_values(f)
                if unused_return:
                    for node in unused_return:
                        info: DETECTOR_INFO = [f, " ignores return value by ", node, "\n"]

                        res = self.generate_result(info)

                        results.append(res)

        return results



from pathlib import Path
import shutil
import sysconfig

from slither.utils.colors import yellow, green, red


def check_path_config(name: str) -> tuple[bool, Path | None, list[Path]]:
    """
    Check if a given Python binary/script is in PATH.
    :return: Returns if the binary on PATH corresponds to this installation,
             its path (if present), and a list of possible paths where this
             binary might be found.
    """
    binary_path = shutil.which(name)
    possible_paths = []

    for scheme in sysconfig.get_scheme_names():
        script_path = Path(sysconfig.get_path("scripts", scheme))
        purelib_path = Path(sysconfig.get_path("purelib", scheme))
        script_binary_path = shutil.which(name, path=script_path)
        if script_binary_path is not None:
            possible_paths.append((script_path, purelib_path))

    binary_here = False
    if binary_path is not None:
        binary_path = Path(binary_path).resolve()
        this_code = Path(__file__).resolve()
        this_binary = list(filter(lambda x: this_code.is_relative_to(x[1]), possible_paths))
        binary_here = len(this_binary) > 0 and all(
            binary_path.is_relative_to(script) for script, _ in this_binary
        )

    return binary_here, binary_path, list(set(script for script, _ in possible_paths))


def check_slither_path(**_kwargs) -> None:
    binary_here, binary_path, possible_paths = check_path_config("slither")
    show_paths = False

    if binary_path:
        print(green(f"`slither` found in PATH at `{binary_path}`."))
        if binary_here:
            print(green("Its location matches this slither-doctor installation."))
        else:
            print(
                yellow(
                    "This path does not correspond to this slither-doctor installation.\n"
                    + "Double-check the order of directories in PATH if you have several Slither installations."
                )
            )
            show_paths = True
    else:
        print(red("`slither` was not found in PATH."))
        show_paths = True

    if show_paths:
        print()
        print("Consider adding one of the following directories to PATH:")
        for path in possible_paths:
            print(f"  * {path}")



from slither.core.expressions import Literal
from slither.core.variables.variable import Variable
from slither.tools.mutator.mutators.abstract_mutator import AbstractMutator
from slither.tools.mutator.utils.patch import create_patch_with_line
from slither.core.solidity_types import ElementaryType

literal_replacements = []


class LIR(AbstractMutator):
    NAME = "LIR"
    HELP = "Literal Integer Replacement"

    def _mutate(self) -> dict:
        result: dict = {}
        variable: Variable

        # Create fault for state variables declaration
        for variable in self.contract.state_variables_declared:
            if variable.initialized:
                # Cannot remove the initialization of constant variables
                if variable.is_constant:
                    continue

                if isinstance(variable.expression, Literal):
                    if isinstance(variable.type, ElementaryType):
                        literal_replacements.append(variable.type.min)  # append data type min value
                        literal_replacements.append(variable.type.max)  # append data type max value
                        if str(variable.type).startswith("uint"):
                            literal_replacements.append("1")
                        elif str(variable.type).startswith("int"):
                            literal_replacements.append("-1")
                    # Get the string
                    start = variable.source_mapping.start
                    stop = start + variable.source_mapping.length
                    old_str = variable.source_mapping.content
                    line_no = variable.node_initialization.source_mapping.lines
                    if line_no[0] not in self.dont_mutate_line:
                        for value in literal_replacements:
                            old_value = old_str[old_str.find("=") + 1 :].strip()
                            if old_value != value:
                                new_str = f"{old_str.split('=')[0]}= {value}"
                                create_patch_with_line(
                                    result,
                                    self.in_file,
                                    start,
                                    stop,
                                    old_str,
                                    new_str,
                                    line_no[0],
                                )

        for function in self.contract.functions_and_modifiers_declared:
            if not self.should_mutate_function(function):
                continue
            for variable in function.local_variables:
                if variable.initialized and isinstance(variable.expression, Literal):
                    if isinstance(variable.type, ElementaryType):
                        literal_replacements.append(variable.type.min)
                        literal_replacements.append(variable.type.max)
                        if str(variable.type).startswith("uint"):
                            literal_replacements.append("1")
                        elif str(variable.type).startswith("int"):
                            literal_replacements.append("-1")
                    start = variable.source_mapping.start
                    stop = start + variable.source_mapping.length
                    old_str = variable.source_mapping.content
                    line_no = variable.source_mapping.lines
                    if line_no[0] not in self.dont_mutate_line:
                        for new_value in literal_replacements:
                            old_value = old_str[old_str.find("=") + 1 :].strip()
                            if old_value != new_value:
                                new_str = f"{old_str.split('=')[0]}= {new_value}"
                                create_patch_with_line(
                                    result,
                                    self.in_file,
                                    start,
                                    stop,
                                    old_str,
                                    new_str,
                                    line_no[0],
                                )
        return result



from typing import TYPE_CHECKING, Optional

from slither.core.declarations.custom_error import CustomError
from slither.core.declarations.custom_error_contract import CustomErrorContract
from slither.core.declarations.custom_error_top_level import CustomErrorTopLevel
from slither.core.variables.local_variable import LocalVariable
from slither.solc_parsing.declarations.caller_context import CallerContextExpression
from slither.solc_parsing.variables.local_variable import LocalVariableSolc

if TYPE_CHECKING:
    from slither.solc_parsing.slither_compilation_unit_solc import SlitherCompilationUnitSolc
    from slither.core.compilation_unit import SlitherCompilationUnit
    from slither.solc_parsing.declarations.contract import ContractSolc


# Part of the code was copied from the function parsing
# In the long term we should refactor these two classes to merge the duplicated code


class CustomErrorSolc(CallerContextExpression):
    def __init__(
        self,
        custom_error: CustomError,
        custom_error_data: dict,
        contract_parser: Optional["ContractSolc"],
        slither_parser: "SlitherCompilationUnitSolc",
    ) -> None:
        self._slither_parser: SlitherCompilationUnitSolc = slither_parser
        self._custom_error = custom_error
        custom_error.name = custom_error_data["name"]
        self._contract_parser = contract_parser
        self._params_was_analyzed = False

        if not self._slither_parser.is_compact_ast:
            custom_error_data = custom_error_data["attributes"]
        self._custom_error_data = custom_error_data

    def analyze_params(self) -> None:
        # Can be re-analyzed due to inheritance
        if self._params_was_analyzed:
            return

        self._params_was_analyzed = True

        if self._slither_parser.is_compact_ast:
            params = self._custom_error_data["parameters"]
        else:
            children = self._custom_error_data[self.get_children("children")]
            # It uses to be
            # params = children[0]
            # returns = children[1]
            # But from Solidity 0.6.3 to 0.6.10 (included)
            # Comment above a function might be added in the children
            child_iter = iter(
                [child for child in children if child[self.get_key()] == "ParameterList"]
            )
            params = next(child_iter)

        if params:
            self._parse_params(params)

    @property
    def contract_parser(self) -> Optional["ContractSolc"]:
        return self._contract_parser

    @property
    def is_compact_ast(self) -> bool:
        return self._slither_parser.is_compact_ast

    def get_key(self) -> str:
        return self._slither_parser.get_key()

    def get_children(self, key: str) -> str:
        if self._slither_parser.is_compact_ast:
            return key
        return "children"

    def _parse_params(self, params: dict) -> None:
        assert params[self.get_key()] == "ParameterList"

        if self._slither_parser.is_compact_ast:
            params = params["parameters"]
        else:
            params = params[self.get_children("children")]

        for param in params:
            assert param[self.get_key()] == "VariableDeclaration"
            local_var = self._add_param(param)
            self._custom_error.add_parameters(local_var.underlying_variable)
        self._custom_error.set_solidity_sig()

    def _add_param(self, param: dict) -> LocalVariableSolc:
        local_var = LocalVariable()
        local_var.set_offset(param["src"], self._slither_parser.compilation_unit)

        local_var_parser = LocalVariableSolc(local_var, param)

        if isinstance(self._custom_error, CustomErrorTopLevel):
            local_var_parser.analyze(self)
        else:
            assert isinstance(self._custom_error, CustomErrorContract)
            local_var_parser.analyze(self)

        # see https://solidity.readthedocs.io/en/v0.4.24/types.html?highlight=storage%20location#data-location
        if local_var.location == "default":
            local_var.set_location("memory")

        return local_var_parser

    @property
    def underlying_custom_error(self) -> CustomError:
        return self._custom_error

    @property
    def slither_parser(self) -> "SlitherCompilationUnitSolc":
        return self._slither_parser

    @property
    def compilation_unit(self) -> "SlitherCompilationUnit":
        return self._custom_error.compilation_unit



from slither.core.compilation_unit import SlitherCompilationUnit
from slither.formatters.utils.patches import create_patch


def custom_format(compilation_unit: SlitherCompilationUnit, result: dict) -> None:
    elements = result["elements"]
    for element in elements:
        if element["type"] == "variable":
            _patch(
                compilation_unit,
                result,
                element["source_mapping"]["filename_absolute"],
                element["source_mapping"]["start"],
            )


def _patch(
    compilation_unit: SlitherCompilationUnit, result: dict, in_file: str, modify_loc_start: int
) -> None:
    in_file_str = compilation_unit.core.source_code[in_file].encode("utf8")
    old_str_of_interest = in_file_str[modify_loc_start:]
    old_str = (
        old_str_of_interest.decode("utf8").partition(";")[0]
        + old_str_of_interest.decode("utf8").partition(";")[1]
    )

    create_patch(
        result,
        in_file,
        int(modify_loc_start),
        # Remove the entire declaration until the semicolon
        int(modify_loc_start + len(old_str_of_interest.decode("utf8").partition(";")[0]) + 1),
        old_str,
        "",
    )



from pathlib import Path
from crytic_compile import CryticCompile
from crytic_compile.platform.solc_standard_json import SolcStandardJson

from slither import Slither

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"
SCOPE_RENAMING_TEST_DATA_DIR = Path(TEST_DATA_DIR, "scope_with_renaming")


# https://github.com/crytic/slither/issues/2454
def test_find_variable_scope_with_renaming(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.24")
    standard_json = SolcStandardJson()
    for source_file in SCOPE_RENAMING_TEST_DATA_DIR.rglob("**/*.sol"):
        standard_json.add_source_file(Path(source_file).as_posix())
    compilation = CryticCompile(standard_json, solc=solc_path)
    Slither(compilation, disallow_partial=True)



from typing import TYPE_CHECKING

from slither.exceptions import SlitherException
from slither.utils.integer_conversion import convert_string_to_fraction


if TYPE_CHECKING:
    from slither.core.declarations import Contract, Function


def convert_subdenomination(value: str, sub: str) -> int:
    decimal_value = convert_string_to_fraction(value)
    if sub == "wei":
        return int(decimal_value)
    if sub == "gwei":
        return int(decimal_value * int(1e9))
    if sub == "szabo":
        return int(decimal_value * int(1e12))
    if sub == "finney":
        return int(decimal_value * int(1e15))
    if sub == "ether":
        return int(decimal_value * int(1e18))
    if sub == "seconds":
        return int(decimal_value)
    if sub == "minutes":
        return int(decimal_value * 60)
    if sub == "hours":
        return int(decimal_value * 60 * 60)
    if sub == "days":
        return int(decimal_value * 60 * 60 * 24)
    if sub == "weeks":
        return int(decimal_value * 60 * 60 * 24 * 7)
    if sub == "years":
        return int(decimal_value * 60 * 60 * 24 * 7 * 365)

    raise SlitherException(f"Subdemonination conversion impossible {decimal_value} {sub}")


# Number of unchecked arithmetic operation needed to be interesting
THRESHOLD_ARITHMETIC_USAGE = 3


def _unchecked_arithemtic_usage(function: "Function") -> bool:
    """
    Check if the function has more than THRESHOLD_ARITHMETIC_USAGE unchecked arithmetic operation

    Args:
        function:

    Returns:

    """

    from slither.slithir.operations import Binary

    score = 0
    for node in function.nodes:
        if not node.scope.is_checked:
            for ir in node.irs:
                if isinstance(ir, Binary):
                    score += 1
                    if score >= THRESHOLD_ARITHMETIC_USAGE:
                        return True
    return False


def unchecked_arithemtic_usage(contract: "Contract") -> list["Function"]:
    """
    Return the list of function with some unchecked arithmetics

    Args:
        contract:

    Returns:

    """

    from slither.core.declarations import Function

    ret: list[Function] = []
    for function in contract.all_functions_called:
        if isinstance(function, Function) and _unchecked_arithemtic_usage(function):
            ret.append(function)
    return ret



from fractions import Fraction
from Crypto.Hash import keccak

from slither.core import expressions
from slither.core.expressions import (
    BinaryOperationType,
    Literal,
    UnaryOperationType,
    Identifier,
    BinaryOperation,
    UnaryOperation,
    TupleExpression,
    TypeConversion,
    CallExpression,
    MemberAccess,
)
from slither.core.expressions.elementary_type_name_expression import ElementaryTypeNameExpression
from slither.core.variables import Variable
from slither.utils.integer_conversion import convert_string_to_fraction, convert_string_to_int
from slither.visitors.expression.expression import ExpressionVisitor
from slither.core.solidity_types.elementary_type import ElementaryType


class NotConstant(Exception):
    pass


KEY = "ConstantFolding"

CONSTANT_TYPES_OPERATIONS = (
    Literal
    | BinaryOperation
    | UnaryOperation
    | Identifier
    | TupleExpression
    | TypeConversion
    | MemberAccess
)


def get_val(expression: CONSTANT_TYPES_OPERATIONS) -> bool | int | Fraction | str:
    val = expression.context[KEY]
    # we delete the item to reduce memory use
    del expression.context[KEY]
    return val


def set_val(expression: CONSTANT_TYPES_OPERATIONS, val: bool | int | Fraction | str) -> None:
    expression.context[KEY] = val


class ConstantFolding(ExpressionVisitor):
    def __init__(
        self, expression: CONSTANT_TYPES_OPERATIONS, custom_type: str | ElementaryType
    ) -> None:
        if isinstance(custom_type, str):
            custom_type = ElementaryType(custom_type)
        self._type: ElementaryType = custom_type
        super().__init__(expression)

    @property
    def expression(self) -> CONSTANT_TYPES_OPERATIONS:
        # We make the assumption that the expression is always a CONSTANT_TYPES_OPERATIONS
        # Other expression are not supported for constant unfolding
        return self._expression  # type: ignore

    def result(self) -> "Literal":
        value = get_val(self.expression)
        if isinstance(value, Fraction):
            value = int(value)
            # emulate 256-bit wrapping
            if str(self._type).startswith("uint"):
                value = value & (2**256 - 1)
        return Literal(value, self._type)

    def _post_identifier(self, expression: Identifier) -> None:
        from slither.core.declarations.solidity_variables import SolidityFunction
        from slither.core.declarations.enum import Enum
        from slither.core.solidity_types.type_alias import TypeAlias
        from slither.core.declarations.contract import Contract

        if isinstance(expression.value, Variable):
            if expression.value.is_constant:
                expr = expression.value.expression
                # assumption that we won't have infinite loop
                # Everything outside of literal
                if isinstance(
                    expr,
                    (
                        BinaryOperation,
                        UnaryOperation,
                        Identifier,
                        TupleExpression,
                        TypeConversion,
                        MemberAccess,
                    ),
                ):
                    cf = ConstantFolding(expr, self._type)
                    expr = cf.result()
                assert isinstance(expr, Literal)
                set_val(expression, convert_string_to_int(expr.converted_value))
            else:
                raise NotConstant
        elif isinstance(expression.value, SolidityFunction):
            set_val(expression, expression.value)
        else:
            # Enum: We don't want to raise an error for a direct access to an Enum as they can be converted to a constant value
            # We can't handle it here because we don't have the field accessed so we do it in _post_member_access
            # TypeAlias: Support when a .wrap() is done with a constant
            # Contract: Support when a constatn is use from a different contract
            if not isinstance(expression.value, (Enum, TypeAlias, Contract)):
                raise NotConstant

    def _post_binary_operation(self, expression: BinaryOperation) -> None:
        expression_left = expression.expression_left
        expression_right = expression.expression_right
        if not isinstance(
            expression_left,
            (
                Literal,
                BinaryOperation,
                UnaryOperation,
                Identifier,
                TupleExpression,
                TypeConversion,
                MemberAccess,
            ),
        ):
            raise NotConstant
        if not isinstance(
            expression_right,
            (
                Literal,
                BinaryOperation,
                UnaryOperation,
                Identifier,
                TupleExpression,
                TypeConversion,
                MemberAccess,
            ),
        ):
            raise NotConstant
        left = get_val(expression_left)
        right = get_val(expression_right)

        if (
            expression.type == BinaryOperationType.POWER
            and isinstance(left, (int, Fraction))
            and isinstance(right, (int, Fraction))
        ):
            set_val(expression, left**right)  # type: ignore
        elif (
            expression.type == BinaryOperationType.MULTIPLICATION
            and isinstance(left, (int, Fraction))
            and isinstance(right, (int, Fraction))
        ):
            set_val(expression, left * right)
        elif (
            expression.type == BinaryOperationType.DIVISION
            and isinstance(left, (int, Fraction))
            and isinstance(right, (int, Fraction))
        ):
            # TODO: maybe check for right + left to be int to use // ?
            set_val(expression, left // right if isinstance(right, int) else left / right)
        elif (
            expression.type == BinaryOperationType.MODULO
            and isinstance(left, (int, Fraction))
            and isinstance(right, (int, Fraction))
        ):
            set_val(expression, left % right)
        elif (
            expression.type == BinaryOperationType.ADDITION
            and isinstance(left, (int, Fraction))
            and isinstance(right, (int, Fraction))
        ):
            set_val(expression, left + right)
        elif (
            expression.type == BinaryOperationType.SUBTRACTION
            and isinstance(left, (int, Fraction))
            and isinstance(right, (int, Fraction))
        ):
            set_val(expression, left - right)
        # Convert to int for operations not supported by Fraction
        elif expression.type == BinaryOperationType.LEFT_SHIFT:
            set_val(expression, int(left) << int(right))
        elif expression.type == BinaryOperationType.RIGHT_SHIFT:
            set_val(expression, int(left) >> int(right))
        elif expression.type == BinaryOperationType.AND:
            set_val(expression, int(left) & int(right))
        elif expression.type == BinaryOperationType.CARET:
            set_val(expression, int(left) ^ int(right))
        elif expression.type == BinaryOperationType.OR:
            set_val(expression, int(left) | int(right))
        elif expression.type == BinaryOperationType.LESS:
            set_val(expression, int(left) < int(right))
        elif expression.type == BinaryOperationType.LESS_EQUAL:
            set_val(expression, int(left) <= int(right))
        elif expression.type == BinaryOperationType.GREATER:
            set_val(expression, int(left) > int(right))
        elif expression.type == BinaryOperationType.GREATER_EQUAL:
            set_val(expression, int(left) >= int(right))
        elif expression.type == BinaryOperationType.EQUAL:
            set_val(expression, int(left) == int(right))
        elif expression.type == BinaryOperationType.NOT_EQUAL:
            set_val(expression, int(left) != int(right))
        # Convert boolean literals from string to bool
        elif expression.type == BinaryOperationType.ANDAND:
            set_val(expression, left == "true" and right == "true")
        elif expression.type == BinaryOperationType.OROR:
            set_val(expression, left == "true" or right == "true")
        else:
            raise NotConstant

    def _post_unary_operation(self, expression: UnaryOperation) -> None:
        # Case of uint a = -7; uint[-a] arr;
        if expression.type == UnaryOperationType.MINUS_PRE:
            expr = expression.expression
            # Everything outside of literal
            if isinstance(
                expr, (BinaryOperation, UnaryOperation, Identifier, TupleExpression, TypeConversion)
            ):
                cf = ConstantFolding(expr, self._type)
                expr = cf.result()
            assert isinstance(expr, Literal)
            set_val(expression, -convert_string_to_fraction(expr.converted_value))
        else:
            raise NotConstant

    def _post_literal(self, expression: Literal) -> None:
        if str(expression.type) == "bool":
            set_val(expression, expression.converted_value)
        elif str(expression.type) == "string":
            set_val(expression, expression.converted_value)
        else:
            try:
                set_val(expression, convert_string_to_fraction(expression.converted_value))
            except ValueError as e:
                raise NotConstant from e

    def _post_assignement_operation(self, expression: expressions.AssignmentOperation) -> None:
        raise NotConstant

    def _post_call_expression(self, expression: expressions.CallExpression) -> None:
        from slither.core.declarations.solidity_variables import SolidityFunction
        from slither.core.declarations.enum import Enum
        from slither.core.solidity_types import TypeAlias

        if (
            isinstance(expression.called, Identifier)
            and expression.called.value == SolidityFunction("type()")
            and len(expression.arguments) == 1
            and (
                isinstance(expression.arguments[0], ElementaryTypeNameExpression)
                or (
                    isinstance(expression.arguments[0], Identifier)
                    and isinstance(expression.arguments[0].value, Enum)
                )
            )
        ):
            # Returning early to support type(ElemType).max/min or type(MyEnum).max/min
            return
        if (
            isinstance(expression.called.expression, Identifier)
            and isinstance(expression.called.expression.value, TypeAlias)
            and isinstance(expression.called, MemberAccess)
            and expression.called.member_name == "wrap"
            and len(expression.arguments) == 1
        ):
            # Handle constants in .wrap of user defined type
            set_val(expression, get_val(expression.arguments[0]))
            return

        called = get_val(expression.called)
        args = [get_val(arg) for arg in expression.arguments]
        if called.name == "keccak256(bytes)":
            digest = keccak.new(digest_bits=256)
            digest.update(str(args[0]).encode("utf8"))
            set_val(expression, digest.digest())
        else:
            raise NotConstant

    def _post_conditional_expression(self, expression: expressions.ConditionalExpression) -> None:
        raise NotConstant

    def _post_elementary_type_name_expression(
        self, expression: expressions.ElementaryTypeNameExpression
    ) -> None:
        # We don't have to raise an exception to support type(uint112).max or similar
        pass

    def _post_index_access(self, expression: expressions.IndexAccess) -> None:
        raise NotConstant

    def _post_member_access(self, expression: expressions.MemberAccess) -> None:
        from slither.core.declarations import (
            SolidityFunction,
            Contract,
            EnumContract,
            EnumTopLevel,
            Enum,
        )
        from slither.core.solidity_types import UserDefinedType, TypeAlias

        if isinstance(expression.expression, CallExpression) and expression.member_name in [
            "min",
            "max",
        ]:
            if isinstance(expression.expression.called, Identifier):
                if expression.expression.called.value == SolidityFunction("type()"):
                    assert len(expression.expression.arguments) == 1
                    type_expression_found = expression.expression.arguments[0]
                    type_found: ElementaryType | UserDefinedType
                    if isinstance(type_expression_found, ElementaryTypeNameExpression):
                        type_expression_found_type = type_expression_found.type
                        assert isinstance(type_expression_found_type, ElementaryType)
                        type_found = type_expression_found_type
                        value = (
                            type_found.max if expression.member_name == "max" else type_found.min
                        )
                        set_val(expression, value)
                        return
                    # type(enum).max/min
                    # Case when enum is in another contract e.g. type(C.E).max
                    if isinstance(type_expression_found, MemberAccess):
                        contract = type_expression_found.expression.value
                        assert isinstance(contract, Contract)
                        for enum in contract.enums:
                            if enum.name == type_expression_found.member_name:
                                type_found_in_expression = enum
                                type_found = UserDefinedType(enum)
                                break
                    else:
                        assert isinstance(type_expression_found, Identifier)
                        type_found_in_expression = type_expression_found.value
                        assert isinstance(type_found_in_expression, (EnumContract, EnumTopLevel))
                        type_found = UserDefinedType(type_found_in_expression)
                    value = (
                        type_found_in_expression.max
                        if expression.member_name == "max"
                        else type_found_in_expression.min
                    )
                    set_val(expression, value)
                    return
        elif isinstance(expression.expression, Identifier) and isinstance(
            expression.expression.value, Enum
        ):
            # Handle direct access to enum field
            set_val(expression, expression.expression.value.values.index(expression.member_name))
            return
        elif isinstance(expression.expression, Identifier) and isinstance(
            expression.expression.value, TypeAlias
        ):
            # User defined type .wrap call handled in _post_call_expression
            return
        elif (
            isinstance(expression.expression, TypeConversion)
            and expression.expression.type == ElementaryType("address")
            and expression.member_name in ["balance", "code", "codehash"]
        ):
            # We need to raise NotConstant for these case here otherwise expression.expression.value would crash in the following condition
            # because TypeConversion does not have a value. See https://github.com/crytic/slither/issues/2717
            raise NotConstant
        elif (
            isinstance(expression.expression.value, Contract)
            and expression.member_name in expression.expression.value.variables_as_dict
            and expression.expression.value.variables_as_dict[expression.member_name].is_constant
        ):
            # Handles when a constant is accessed on another contract
            variables = expression.expression.value.variables_as_dict
            if isinstance(variables[expression.member_name].expression, MemberAccess):
                self._post_member_access(variables[expression.member_name].expression)
                set_val(expression, get_val(variables[expression.member_name].expression))
                return

            # If the variable is a Literal we convert its value to int
            if isinstance(variables[expression.member_name].expression, Literal):
                value = convert_string_to_int(
                    variables[expression.member_name].expression.converted_value
                )
            # If the variable is a UnaryOperation we need convert its value to int
            # and replacing possible spaces
            elif isinstance(variables[expression.member_name].expression, UnaryOperation):
                value = convert_string_to_int(
                    str(variables[expression.member_name].expression).replace(" ", "")
                )
            else:
                value = variables[expression.member_name].expression

            set_val(expression, value)
            return

        raise NotConstant

    def _post_new_array(self, expression: expressions.NewArray) -> None:
        raise NotConstant

    def _post_new_contract(self, expression: expressions.NewContract) -> None:
        raise NotConstant

    def _post_new_elementary_type(self, expression: expressions.NewElementaryType) -> None:
        raise NotConstant

    def _post_tuple_expression(self, expression: expressions.TupleExpression) -> None:
        if expression.expressions:
            if len(expression.expressions) == 1:
                first_expr = expression.expressions[0]
                if not isinstance(
                    first_expr,
                    (
                        Literal,
                        BinaryOperation,
                        UnaryOperation,
                        Identifier,
                        TupleExpression,
                        TypeConversion,
                    ),
                ):
                    raise NotConstant
                cf = ConstantFolding(first_expr, self._type)
                expr = cf.result()
                assert isinstance(expr, Literal)
                set_val(expression, convert_string_to_fraction(expr.converted_value))
                return
        raise NotConstant

    def _post_type_conversion(self, expression: expressions.TypeConversion) -> None:
        expr = expression.expression
        if not isinstance(
            expr,
            (
                Literal,
                BinaryOperation,
                UnaryOperation,
                Identifier,
                TupleExpression,
                TypeConversion,
                CallExpression,
                MemberAccess,
            ),
        ):
            raise NotConstant
        cf = ConstantFolding(expr, self._type)
        expr = cf.result()
        assert isinstance(expr, Literal)
        if str(expression.type).startswith("uint") and isinstance(expr.value, bytes):
            value = int.from_bytes(expr.value, "big")
        elif str(expression.type).startswith("byte") and isinstance(expr.value, int):
            value = int.to_bytes(expr.value, 32, "big")
        elif str(expression.type).startswith("byte") and isinstance(expr.value, str):
            value = expr.value
        else:
            value = convert_string_to_fraction(expr.converted_value)
        set_val(expression, value)



from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import Nop
from slither.utils.output import Output


class VoidConstructor(AbstractDetector):
    ARGUMENT = "void-cst"
    HELP = "Constructor called not implemented"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#void-constructor"

    WIKI_TITLE = "Void constructor"
    WIKI_DESCRIPTION = "Detect the call to a constructor that is not implemented"
    WIKI_RECOMMENDATION = "Remove the constructor call."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract A{}
contract B is A{
    constructor() public A(){}
}
```
When reading `B`'s constructor definition, we might assume that `A()` initiates the contract, but no code is executed."""
    # endregion wiki_exploit_scenario

    def _detect(self) -> list[Output]:
        """"""
        results = []
        for c in self.contracts:
            cst = c.constructor
            if cst:
                for constructor_call in cst.explicit_base_constructor_calls_statements:
                    for node in constructor_call.nodes:
                        if any(isinstance(ir, Nop) for ir in node.irs):
                            info: DETECTOR_INFO = ["Void constructor called in ", cst, ":\n"]
                            info += ["\t- ", node, "\n"]

                            res = self.generate_result(info)

                            results.append(res)
        return results



from slither.core.solidity_types.elementary_type import ElementaryType
from slither.slithir.operations.call import Call
from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.utils.utils import is_valid_lvalue


class NewElementaryType(Call, OperationWithLValue):
    def __init__(self, new_type, lvalue):
        assert isinstance(new_type, ElementaryType)
        assert is_valid_lvalue(lvalue)
        super().__init__()
        self._type = new_type
        self._lvalue = lvalue

    @property
    def type(self):
        return self._type

    @property
    def read(self):
        return list(self.arguments)

    def __str__(self):
        args = [str(a) for a in self.arguments]

        return f"{self.lvalue} = new {self._type}({','.join(args)})"



from .summary.function import FunctionSummary
from .summary.contract import ContractSummary
from .summary.loc import LocPrinter
from .inheritance.inheritance import PrinterInheritance
from .inheritance.inheritance_graph import PrinterInheritanceGraph
from .call.call_graph import PrinterCallGraph
from .functions.authorization import PrinterWrittenVariablesAndAuthorization
from .summary.slithir import PrinterSlithIR
from .summary.slithir_ssa import PrinterSlithIRSSA
from .summary.human_summary import PrinterHumanSummary
from .summary.ck import CK
from .summary.halstead import Halstead
from .functions.cfg import CFG
from .summary.function_ids import FunctionIds
from .summary.variable_order import VariableOrder
from .summary.data_depenency import DataDependency
from .summary.modifier_calls import Modifiers
from .summary.require_calls import RequireOrAssert
from .summary.constructor_calls import ConstructorPrinter
from .guidance.echidna import Echidna
from .summary.evm import PrinterEVM
from .summary.when_not_paused import PrinterWhenNotPaused
from .summary.declaration import Declaration
from .functions.dominator import Dominator
from .summary.martin import Martin
from .summary.cheatcodes import CheatcodePrinter
from .summary.entry_points import PrinterEntryPoints



import logging
import re
from pathlib import Path

from slither.tools.properties.addresses.address import Addresses
from slither.tools.properties.properties.properties import (
    PropertyReturn,
    Property,
    PropertyCaller,
)
from slither.tools.properties.utils import write_file

PATTERN_TRUFFLE_MIGRATION = re.compile("^[0-9]*_")
logger = logging.getLogger("Slither")


def _extract_caller(p: PropertyCaller) -> list[str]:
    if p == PropertyCaller.OWNER:
        return ["owner"]
    if p == PropertyCaller.SENDER:
        return ["user"]
    if p == PropertyCaller.ATTACKER:
        return ["attacker"]
    if p == PropertyCaller.ALL:
        return ["owner", "user", "attacker"]
    assert p == PropertyCaller.ANY
    return ["user"]


def _helpers() -> str:
    """
    Generate two functions:
    - catchRevertThrowReturnFalse: check if the call revert/throw or return false
    - catchRevertThrow: check if the call revert/throw
    :return:
    """
    return """
    async function catchRevertThrowReturnFalse(promise) {
    try {
        const ret = await promise;
        assert.equal(balance.valueOf(), false, "Expected revert/throw/or return false");
    } catch (error) {
        // Not considered: 'out of gas', 'invalid JUMP'
        if (!error.message.includes("revert")){
            if (!error.message.includes("invalid opcode")){
                assert(false, "Expected revert/throw/or return false");
            }
        }
        return;
    }
};

async function catchRevertThrow(promise) {
    try {
        await promise;
    } catch (error) {
        // Not considered: 'out of gas', 'invalid JUMP'
        if (!error.message.includes("revert")){
            if (!error.message.includes("invalid opcode")){
                assert(false, "Expected revert/throw/or return false");
            }
        }
        return;
    }
    assert(false, "Expected revert/throw/or return false");
};
"""


def generate_unit_test(
    test_contract: str,
    filename: str,
    unit_tests: list[Property],
    output_dir: Path,
    addresses: Addresses,
    assert_message: str = "",
) -> Path:
    """
    Generate unit tests files
    :param test_contract:
    :param filename:
    :param unit_tests:
    :param output_dir:
    :param addresses:
    :param assert_message:
    :return:
    """
    content = f'{test_contract} = artifacts.require("{test_contract}");\n\n'

    content += _helpers()

    content += f'contract("{test_contract}", accounts => {{\n'

    content += f'\tlet owner = "{addresses.owner}";\n'
    content += f'\tlet user = "{addresses.user}";\n'
    content += f'\tlet attacker = "{addresses.attacker}";\n'
    for unit_test in unit_tests:
        content += f'\tit("{unit_test.description}", async () => {{\n'
        content += f"\t\tlet instance = await {test_contract}.deployed();\n"
        callers = _extract_caller(unit_test.caller)
        if unit_test.return_type == PropertyReturn.SUCCESS:
            for caller in callers:
                content += f"\t\tlet test_{caller} = await instance.{unit_test.name[:-2]}.call({{from: {caller}}});\n"
                if assert_message:
                    content += f'\t\tassert.equal(test_{caller}, true, "{assert_message}");\n'
                else:
                    content += f"\t\tassert.equal(test_{caller}, true);\n"
        elif unit_test.return_type == PropertyReturn.FAIL:
            for caller in callers:
                content += f"\t\tlet test_{caller} = await instance.{unit_test.name[:-2]}.call({{from: {caller}}});\n"
                if assert_message:
                    content += f'\t\tassert.equal(test_{caller}, false, "{assert_message}");\n'
                else:
                    content += f"\t\tassert.equal(test_{caller}, false);\n"
        elif unit_test.return_type == PropertyReturn.FAIL_OR_THROW:
            for caller in callers:
                content += f"\t\tawait catchRevertThrowReturnFalse(instance.{unit_test.name[:-2]}.call({{from: {caller}}}));\n"
        elif unit_test.return_type == PropertyReturn.THROW:
            callers = _extract_caller(unit_test.caller)
            for caller in callers:
                content += f"\t\tawait catchRevertThrow(instance.{unit_test.name[:-2]}.call({{from: {caller}}}));\n"
        content += "\t});\n"

    content += "});\n"

    output_dir = Path(output_dir, "test")
    output_dir.mkdir(exist_ok=True)

    output_dir = Path(output_dir, "crytic")
    output_dir.mkdir(exist_ok=True)

    write_file(output_dir, filename, content)
    return output_dir


def generate_migration(test_contract: str, output_dir: Path, owner_address: str) -> None:
    """
    Generate migration file
    :param test_contract:
    :param output_dir:
    :param owner_address:
    :return:
    """
    content = f"""{test_contract} = artifacts.require("{test_contract}");
module.exports = function(deployer) {{
  deployer.deploy({test_contract}, {{from: "{owner_address}"}});
}};
"""

    output_dir = Path(output_dir, "migrations")

    output_dir.mkdir(exist_ok=True)

    migration_files = [
        js_file
        for js_file in output_dir.iterdir()
        if js_file.suffix == ".js" and PATTERN_TRUFFLE_MIGRATION.match(js_file.name)
    ]

    idx = len(migration_files)
    filename = f"{idx + 1}_{test_contract}.js"
    potential_previous_filename = f"{idx}_{test_contract}.js"

    for m in migration_files:
        if m.name == potential_previous_filename:
            write_file(output_dir, potential_previous_filename, content)
            return
        if test_contract in m.name:
            logger.error(f"Potential conflicts with {m.name}")

    write_file(output_dir, filename, content)



from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.core.cfg.node import Node
from slither.core.variables.variable import Variable
from slither.core.expressions import TypeConversion, Literal
from slither.utils.output import Output


class OptimismDeprecation(AbstractDetector):
    ARGUMENT = "optimism-deprecation"
    HELP = "Detect when deprecated Optimism predeploy or function is used."
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#optimism-deprecated-predeploy-or-function"

    WIKI_TITLE = "Optimism deprecated predeploy or function"
    WIKI_DESCRIPTION = "Detect when deprecated Optimism predeploy or function is used."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
interface GasPriceOracle {
    function scalar() external view returns (uint256);
}

contract Test {
    GasPriceOracle constant OPT_GAS = GasPriceOracle(0x420000000000000000000000000000000000000F);

    function a() public {
        OPT_GAS.scalar();
    }
}
```
The call to the `scalar` function of the Optimism GasPriceOracle predeploy always revert.
"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Do not use the deprecated components."

    def _detect(self) -> list[Output]:
        results = []

        deprecated_predeploys = [
            "0x4200000000000000000000000000000000000000",  # LegacyMessagePasser
            "0x4200000000000000000000000000000000000001",  # L1MessageSender
            "0x4200000000000000000000000000000000000002",  # DeployerWhitelist
            "0x4200000000000000000000000000000000000013",  # L1BlockNumber
        ]

        for contract in self.compilation_unit.contracts_derived:
            use_deprecated: list[Node] = []

            for _, ir in contract.all_high_level_calls:
                # To avoid FPs we assume predeploy contracts are always assigned to a constant and typecasted to an interface
                # and we check the target address of a high level call.
                if (
                    isinstance(ir.destination, Variable)
                    and isinstance(ir.destination.expression, TypeConversion)
                    and isinstance(ir.destination.expression.expression, Literal)
                ):
                    if ir.destination.expression.expression.value in deprecated_predeploys:
                        use_deprecated.append(ir.node)

                    if (
                        ir.destination.expression.expression.value
                        == "0x420000000000000000000000000000000000000F"
                        and ir.function_name in ("overhead", "scalar", "getL1GasUsed")
                    ):
                        use_deprecated.append(ir.node)
            # Sort so output is deterministic
            use_deprecated.sort(key=lambda x: (x.node_id, x.function.full_name))
            if len(use_deprecated) > 0:
                info: DETECTOR_INFO = [
                    "A deprecated Optimism predeploy or function is used in the ",
                    contract.name,
                    " contract.\n",
                ]

                for node in use_deprecated:
                    info.extend(["\t - ", node, "\n"])

                res = self.generate_result(info)
                results.append(res)

        return results



"""
EventTopLevel module
"""

from typing import TYPE_CHECKING

from slither.core.declarations.event_top_level import EventTopLevel
from slither.core.variables.event_variable import EventVariable
from slither.core.compilation_unit import SlitherCompilationUnit
from slither.solc_parsing.variables.event_variable import EventVariableSolc
from slither.solc_parsing.declarations.caller_context import CallerContextExpression

if TYPE_CHECKING:
    from slither.solc_parsing.slither_compilation_unit_solc import SlitherCompilationUnitSolc


class EventTopLevelSolc(CallerContextExpression):
    """
    EventTopLevel class
    """

    def __init__(
        self, event: EventTopLevel, event_data: dict, slither_parser: "SlitherCompilationUnitSolc"
    ) -> None:
        self._event = event
        self._slither_parser = slither_parser

        if self.is_compact_ast:
            self._event.name = event_data["name"]
            elems = event_data["parameters"]
            assert elems["nodeType"] == "ParameterList"
            self._elemsNotParsed = elems["parameters"]
        else:
            self._event.name = event_data["attributes"]["name"]
            for elem in event_data["children"]:
                # From Solidity 0.6.3 to 0.6.10 (included)
                # Comment above a event might be added in the children
                # of an event for the legacy ast
                if elem["name"] == "ParameterList":
                    if "children" in elem:
                        self._elemsNotParsed = elem["children"]
                    else:
                        self._elemsNotParsed = []

    def analyze(self) -> None:
        for elem_to_parse in self._elemsNotParsed:
            elem = EventVariable()
            # Todo: check if the source offset is always here
            if "src" in elem_to_parse:
                elem.set_offset(elem_to_parse["src"], self._slither_parser.compilation_unit)
            elem_parser = EventVariableSolc(elem, elem_to_parse)
            elem_parser.analyze(self)

            self._event.elems.append(elem)

        self._elemsNotParsed = []

    @property
    def is_compact_ast(self) -> bool:
        return self._slither_parser.is_compact_ast

    @property
    def compilation_unit(self) -> SlitherCompilationUnit:
        return self._slither_parser.compilation_unit

    def get_key(self) -> str:
        return self._slither_parser.get_key()

    @property
    def slither_parser(self) -> "SlitherCompilationUnitSolc":
        return self._slither_parser

    @property
    def underlying_event(self) -> EventTopLevel:
        return self._event



from slither.core.declarations import Contract
from slither.tools.upgradeability.checks.abstract_checks import (
    CheckClassification,
    AbstractCheck,
    CHECK_INFO,
)
from slither.utils.upgradeability import get_missing_vars
from slither.utils.output import Output


class MissingVariable(AbstractCheck):
    ARGUMENT = "missing-variables"
    IMPACT = CheckClassification.MEDIUM

    HELP = "Variable missing in the v2"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#missing-variables"
    WIKI_TITLE = "Missing variables"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect variables that were present in the original contracts but are not in the updated one.
"""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract V1{
    uint variable1;
    uint variable2;
}

contract V2{
    uint variable1;
}
```
The new version, `V2` does not contain `variable1`.
If a new variable is added in an update of `V2`, this variable will hold the latest value of `variable2` and
will be corrupted.
"""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Do not change the order of the state variables in the updated contract.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True
    REQUIRE_CONTRACT_V2 = True

    def _check(self) -> list[Output]:
        contract1 = self.contract
        contract2 = self.contract_v2

        assert contract2
        missing = get_missing_vars(contract1, contract2)

        results = []
        for variable1 in missing:
            info: CHECK_INFO = ["Variable missing in ", contract2, ": ", variable1, "\n"]
            json = self.generate_result(info)
            results.append(json)

        return results


class DifferentVariableContractProxy(AbstractCheck):
    ARGUMENT = "order-vars-proxy"
    IMPACT = CheckClassification.HIGH

    HELP = "Incorrect vars order with the proxy"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#incorrect-variables-with-the-proxy"
    WIKI_TITLE = "Incorrect variables with the proxy"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect variables that are different between the contract and the proxy.
"""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Contract{
    uint variable1;
}

contract Proxy{
    address variable1;
}
```
`Contract` and `Proxy` do not have the same storage layout. As a result the storage of both contracts can be corrupted.
"""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Avoid variables in the proxy. If a variable is in the proxy, ensure it has the same layout than in the contract.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True
    REQUIRE_PROXY = True

    def _contract1(self) -> Contract:
        return self.contract

    def _contract2(self) -> Contract:
        assert self.proxy
        return self.proxy

    def _check(self) -> list[Output]:
        contract1 = self._contract1()
        contract2 = self._contract2()

        results: list[Output] = []

        def _check_internal(
            contract1: Contract, contract2: Contract, results: list[Output], is_transient: bool
        ):
            if is_transient:
                order1 = contract1.transient_variables_ordered
                order2 = contract2.transient_variables_ordered
            else:
                order1 = contract1.storage_variables_ordered
                order2 = contract2.storage_variables_ordered

            for idx, _ in enumerate(order1):
                if len(order2) <= idx:
                    # Handle by MissingVariable
                    return

                variable1 = order1[idx]
                variable2 = order2[idx]
                if (variable1.name != variable2.name) or (variable1.type != variable2.type):
                    info: CHECK_INFO = [
                        "Different variables between ",
                        contract1,
                        " and ",
                        contract2,
                        "\n",
                    ]
                    info += ["\t ", variable1, "\n"]
                    info += ["\t ", variable2, "\n"]
                    json = self.generate_result(info)
                    results.append(json)

        # Checking state variables with storage location
        _check_internal(contract1, contract2, results, False)
        # Checking state variables with transient location
        _check_internal(contract1, contract2, results, True)

        return results


class DifferentVariableContractNewContract(DifferentVariableContractProxy):
    ARGUMENT = "order-vars-contracts"

    HELP = "Incorrect vars order with the v2"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#incorrect-variables-with-the-v2"
    WIKI_TITLE = "Incorrect variables with the v2"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect variables that are different between the original contract and the updated one.
"""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Contract{
    uint variable1;
}

contract ContractV2{
    address variable1;
}
```
`Contract` and `ContractV2` do not have the same storage layout. As a result the storage of both contracts can be corrupted.
"""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Respect the variable order of the original contract in the updated contract.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True
    REQUIRE_PROXY = False
    REQUIRE_CONTRACT_V2 = True

    def _contract2(self) -> Contract:
        assert self.contract_v2
        return self.contract_v2


class ExtraVariablesProxy(AbstractCheck):
    ARGUMENT = "extra-vars-proxy"
    IMPACT = CheckClassification.MEDIUM

    HELP = "Extra vars in the proxy"
    WIKI = (
        "https://github.com/crytic/slither/wiki/Upgradeability-Checks#extra-variables-in-the-proxy"
    )
    WIKI_TITLE = "Extra variables in the proxy"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect variables that are in the proxy and not in the contract.
"""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Contract{
    uint variable1;
}

contract Proxy{
    uint variable1;
    uint variable2;
}
```
`Proxy` contains additional variables. A future update of `Contract` is likely to corrupt the proxy.
"""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Avoid variables in the proxy. If a variable is in the proxy, ensure it has the same layout than in the contract.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True
    REQUIRE_PROXY = True

    def _contract1(self) -> Contract:
        return self.contract

    def _contract2(self) -> Contract:
        assert self.proxy
        return self.proxy

    def _check(self) -> list[Output]:
        contract1 = self._contract1()
        contract2 = self._contract2()

        results: list[Output] = []

        def _check_internal(
            contract1: Contract, contract2: Contract, results: list[Output], is_transient: bool
        ):
            if is_transient:
                order1 = contract1.transient_variables_ordered
                order2 = contract2.transient_variables_ordered
            else:
                order1 = contract1.storage_variables_ordered
                order2 = contract2.storage_variables_ordered

            if len(order2) <= len(order1):
                return

            idx = len(order1)

            while idx < len(order2):
                variable2 = order2[idx]
                info: CHECK_INFO = ["Extra variables in ", contract2, ": ", variable2, "\n"]
                json = self.generate_result(info)
                results.append(json)
                idx = idx + 1

        # Checking state variables with storage location
        _check_internal(contract1, contract2, results, False)
        # Checking state variables with transient location
        _check_internal(contract1, contract2, results, True)

        return results


class ExtraVariablesNewContract(ExtraVariablesProxy):
    ARGUMENT = "extra-vars-v2"

    HELP = "Extra vars in the v2"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#extra-variables-in-the-v2"
    WIKI_TITLE = "Extra variables in the v2"

    # region wiki_description
    WIKI_DESCRIPTION = """
Show new variables in the updated contract.

This finding does not have an immediate security impact and is informative.
"""
    # endregion wiki_description

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Ensure that all the new variables are expected.
"""
    # endregion wiki_recommendation

    IMPACT = CheckClassification.INFORMATIONAL

    REQUIRE_CONTRACT = True
    REQUIRE_PROXY = False
    REQUIRE_CONTRACT_V2 = True

    def _contract2(self) -> Contract:
        assert self.contract_v2
        return self.contract_v2



from typing import TYPE_CHECKING

from slither.core.expressions.expression import Expression

if TYPE_CHECKING:
    from slither.core.solidity_types.array_type import ArrayType


class NewArray(Expression):
    def __init__(self, array_type: "ArrayType") -> None:
        super().__init__()

        from slither.core.solidity_types.array_type import ArrayType

        assert isinstance(array_type, ArrayType)
        self._array_type = array_type

    @property
    def array_type(self) -> "ArrayType":
        return self._array_type

    def __str__(self):
        return "new " + str(self._array_type)



from slither.core.cfg.node import NodeType
from slither.tools.mutator.utils.patch import create_patch_with_line
from slither.tools.mutator.mutators.abstract_mutator import AbstractMutator


class CR(AbstractMutator):
    NAME = "CR"
    HELP = "Comment Replacement"

    def _mutate(self) -> dict:
        result: dict = {}

        for function in self.contract.functions_and_modifiers_declared:
            if not self.should_mutate_function(function):
                continue
            for node in function.nodes:
                if not self.should_mutate_node(node):
                    continue
                if node.type not in (
                    NodeType.ENTRYPOINT,
                    NodeType.ENDIF,
                    NodeType.ENDLOOP,
                ):
                    # Get the string
                    start = node.source_mapping.start
                    stop = start + node.source_mapping.length
                    old_str = node.source_mapping.content
                    line_no = node.source_mapping.lines
                    new_str = "//" + old_str
                    create_patch_with_line(
                        result,
                        self.in_file,
                        start,
                        stop,
                        old_str,
                        new_str,
                        line_no[0],
                    )
        return result



import re
from slither.core.cfg.node import NodeType
from slither.tools.mutator.utils.patch import create_patch_with_line
from slither.tools.mutator.mutators.abstract_mutator import AbstractMutator
from slither.core.variables.variable import Variable

solidity_rules = [
    "abi\\.encode\\( ==> abi.encodePacked(",
    "abi\\.encodePacked\\( ==> abi.encode(",
    "\\.call([({]) ==> .delegatecall\\1",
    "\\.call([({]) ==> .staticcall\\1",
    "\\.delegatecall([({]) ==> .call\\1",
    "\\.delegatecall([({]) ==> .staticcall\\1",
    "\\.staticcall([({]) ==> .delegatecall\\1",
    "\\.staticcall([({]) ==> .call\\1",
    "^now$ ==> 0",
    "block.timestamp ==> 0",
    "msg.value ==> 0",
    "msg.value ==> 1",
    "(\\s)(wei|gwei) ==> \\1ether",
    "(\\s)(ether|gwei) ==> \\1wei",
    "(\\s)(wei|ether) ==> \\1gwei",
    "(\\s)(minutes|days|hours|weeks) ==> \\1seconds",
    "(\\s)(seconds|days|hours|weeks) ==> \\1minutes",
    "(\\s)(seconds|minutes|hours|weeks) ==> \\1days",
    "(\\s)(seconds|minutes|days|weeks) ==> \\1hours",
    "(\\s)(seconds|minutes|days|hours) ==> \\1weeks",
    "(\\s)(memory) ==> \\1storage",
    "(\\s)(storage) ==> \\1memory",
    "(\\s)(constant) ==> \\1immutable",
    "addmod ==> mulmod",
    "mulmod ==> addmod",
    "msg.sender ==> tx.origin",
    "tx.origin ==> msg.sender",
    "([^u])fixed ==> \\1ufixed",
    "ufixed ==> fixed",
    "(u?)int16 ==> \\1int8",
    "(u?)int32 ==> \\1int16",
    "(u?)int64 ==> \\1int32",
    "(u?)int128 ==> \\1int64",
    "(u?)int256 ==> \\1int128",
    "while ==> if",
]


class SBR(AbstractMutator):
    NAME = "SBR"
    HELP = "Solidity Based Replacement"

    def _mutate(self) -> dict:
        result: dict = {}
        variable: Variable

        for function in self.contract.functions_and_modifiers_declared:
            if not self.should_mutate_function(function):
                continue
            for node in function.nodes:
                if not self.should_mutate_node(node):
                    continue
                if node.type not in (
                    NodeType.ENTRYPOINT,
                    NodeType.ENDIF,
                    NodeType.ENDLOOP,
                ):
                    # Get the string
                    start = node.source_mapping.start
                    stop = start + node.source_mapping.length
                    old_str = node.source_mapping.content
                    line_no = node.source_mapping.lines
                    for value in solidity_rules:
                        left_value = value.split(" ==> ", maxsplit=1)[0]
                        right_value = value.split(" ==> ")[1]
                        if re.search(re.compile(left_value), old_str) is not None:
                            new_str = re.sub(re.compile(left_value), right_value, old_str)
                            create_patch_with_line(
                                result,
                                self.in_file,
                                start,
                                stop,
                                old_str,
                                new_str,
                                line_no[0],
                            )

        for variable in self.contract.state_variables_declared:
            node = variable.node_initialization
            if node:
                start = node.source_mapping.start
                stop = start + node.source_mapping.length
                old_str = node.source_mapping.content
                line_no = node.source_mapping.lines
                for value in solidity_rules:
                    left_value = value.split(" ==> ", maxsplit=1)[0]
                    right_value = value.split(" ==> ")[1]
                    if re.search(re.compile(left_value), old_str) is not None:
                        new_str = re.sub(re.compile(left_value), right_value, old_str)
                        create_patch_with_line(
                            result,
                            self.in_file,
                            start,
                            stop,
                            old_str,
                            new_str,
                            line_no[0],
                        )
        return result



from slither.analyses.data_dependency.data_dependency import is_tainted
from slither.core.cfg.node import Node
from slither.core.declarations.function_contract import FunctionContract
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


def controlled_delegatecall(function: FunctionContract) -> list[Node]:
    ret = []
    for ir in function.low_level_calls:
        if ir.function_name in [
            "delegatecall",
            "callcode",
        ]:
            if is_tainted(ir.destination, function.contract):
                ret.append(ir.node)
    return ret


class ControlledDelegateCall(AbstractDetector):
    ARGUMENT = "controlled-delegatecall"
    HELP = "Controlled delegatecall destination"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#controlled-delegatecall"

    WIKI_TITLE = "Controlled Delegatecall"
    WIKI_DESCRIPTION = "`Delegatecall` or `callcode` to an address controlled by the user."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Delegatecall{
    function delegate(address to, bytes data){
        to.delegatecall(data);
    }
}
```
Bob calls `delegate` and delegates the execution to his malicious contract. As a result, Bob withdraws the funds of the contract and destructs it."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Avoid using `delegatecall`. Use only trusted destinations."

    def _detect(self) -> list[Output]:
        results = []

        for contract in self.compilation_unit.contracts_derived:
            for f in contract.functions:
                # If its an upgradeable proxy, do not report protected function
                # As functions to upgrades the destination lead to too many FPs
                if contract.is_upgradeable_proxy and f.is_protected():
                    continue
                nodes = controlled_delegatecall(f)
                if nodes:
                    func_info: DETECTOR_INFO = [
                        f,
                        " uses delegatecall to a input-controlled function id\n",
                    ]

                    for node in nodes:
                        node_info: DETECTOR_INFO = func_info + ["\t- ", node, "\n"]
                        res = self.generate_result(node_info)
                        results.append(res)

        return results



from typing import TYPE_CHECKING

from slither.slithir.variables.variable import SlithIRVariable

if TYPE_CHECKING:
    from slither.core.cfg.node import Node


class TupleVariable(SlithIRVariable):
    def __init__(self, node: "Node", index: int | None = None) -> None:
        super().__init__()
        if index is None:
            self._index = node.compilation_unit.counter_slithir_tuple
            node.compilation_unit.counter_slithir_tuple += 1
        else:
            self._index = index

        self._node = node

    @property
    def node(self) -> "Node":
        return self._node

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, idx):
        self._index = idx

    @property
    def name(self) -> str:
        return f"TUPLE_{self.index}"

    def __str__(self) -> str:
        return self.name



"""
Module detecting usage of low level calls
"""

from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import LowLevelCall
from slither.core.cfg.node import Node
from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.utils.output import Output


class LowLevelCalls(AbstractDetector):
    """
    Detect usage of low level calls
    """

    ARGUMENT = "low-level-calls"
    HELP = "Low level calls"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#low-level-calls"

    WIKI_TITLE = "Low-level calls"
    WIKI_DESCRIPTION = "The use of low-level calls is error-prone. Low-level calls do not check for [code existence](https://solidity.readthedocs.io/en/v0.4.25/control-structures.html#error-handling-assert-require-revert-and-exceptions) or call success."
    WIKI_RECOMMENDATION = "Avoid low-level calls. Check the call success. If the call is meant for a contract, check for code existence."

    @staticmethod
    def _contains_low_level_calls(node: Node) -> bool:
        """
             Check if the node contains Low Level Calls
        Returns:
            (bool)
        """
        return any(isinstance(ir, LowLevelCall) for ir in node.irs)

    def detect_low_level_calls(
        self, contract: Contract
    ) -> list[tuple[FunctionContract, list[Node]]]:
        ret = []
        for f in [f for f in contract.functions if contract == f.contract_declarer]:
            low_level_nodes = [ir.node for ir in f.low_level_calls]
            if low_level_nodes:
                ret.append((f, low_level_nodes))
        return ret

    def _detect(self) -> list[Output]:
        """Detect the functions that use low level calls"""
        results = []
        for c in self.contracts:
            values = self.detect_low_level_calls(c)
            for func, nodes in values:
                info: DETECTOR_INFO = ["Low level call in ", func, ":\n"]

                # sort the nodes to get deterministic results
                nodes.sort(key=lambda x: x.node_id)

                for node in nodes:
                    info += ["\t- ", node, "\n"]

                res = self.generate_result(info)

                results.append(res)

        return results



import sys
from slither.slither import Slither


if len(sys.argv) != 2:
    print("python export_dominator_tree_to_dot.py contract.sol")
    sys.exit(-1)

# Init slither
slither = Slither(sys.argv[1])

for contract in slither.contracts:
    for function in list(contract.functions) + list(contract.modifiers):
        filename = f"{sys.argv[1]}-{contract.name}-{function.full_name}_dom.dot"
        print(f"Export {filename}")
        function.dominator_tree_to_dot(filename)



from typing import Union, TYPE_CHECKING, Any
import math

from slither.core.solidity_types.type import Type
from slither.exceptions import SlitherException

if TYPE_CHECKING:
    from slither.core.declarations.structure import Structure
    from slither.core.declarations.enum import Enum
    from slither.core.declarations.contract import Contract


class UserDefinedType(Type):
    def __init__(self, t: Union["Enum", "Contract", "Structure"]) -> None:
        from slither.core.declarations.structure import Structure
        from slither.core.declarations.enum import Enum
        from slither.core.declarations.contract import Contract

        assert isinstance(t, (Contract, Enum, Structure))
        super().__init__()
        self._type = t

    @property
    def is_dynamic(self) -> bool:
        return False

    @property
    def type(self) -> Union["Contract", "Enum", "Structure"]:
        return self._type

    @property
    def storage_size(self) -> tuple[int, bool]:
        from slither.core.declarations.structure import Structure
        from slither.core.declarations.enum import Enum
        from slither.core.declarations.contract import Contract

        if isinstance(self._type, Contract):
            return 20, False
        if isinstance(self._type, Enum):
            return math.ceil(math.log2(len(self._type.values)) / 8), False
        if isinstance(self._type, Structure):
            # todo there's some duplicate logic here and slither_core, can we refactor this?
            slot = 0
            offset = 0
            for elem in self._type.elems_ordered:
                size, new_slot = elem.type.storage_size
                if new_slot:
                    if offset > 0:
                        slot += 1
                        offset = 0
                elif size + offset > 32:
                    slot += 1
                    offset = 0

                if new_slot:
                    slot += math.ceil(size / 32)
                else:
                    offset += size
            if offset > 0:
                slot += 1
            return slot * 32, True
        to_log = f"{self} does not have storage size"
        raise SlitherException(to_log)

    def __str__(self) -> str:
        from slither.core.declarations.structure_contract import StructureContract
        from slither.core.declarations.enum_contract import EnumContract

        type_used = self.type
        if isinstance(type_used, (EnumContract, StructureContract)):
            return str(type_used.contract) + "." + str(type_used.name)
        return str(type_used.name)

    def __eq__(self, other: Any) -> bool:
        from slither.core.declarations.contract import Contract

        if not isinstance(other, UserDefinedType):
            return False
        if isinstance(self.type, Contract) and isinstance(other.type, Contract):
            return self.type == other.type.name
        return self.type == other.type

    def __hash__(self) -> int:
        return hash(str(self))



"""
Module detecting unused return values from send
"""

from slither.detectors.abstract_detector import DetectorClassification
from slither.detectors.operations.unused_return_values import UnusedReturnValues
from slither.slithir.operations import Send
from slither.slithir.operations.operation import Operation


class UncheckedSend(UnusedReturnValues):
    """
    If the return value of a send is not checked, it might lead to losing ether
    """

    ARGUMENT = "unchecked-send"
    HELP = "Unchecked send"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#unchecked-send"

    WIKI_TITLE = "Unchecked Send"
    WIKI_DESCRIPTION = "The return value of a `send` is not checked."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract MyConc{
    function my_func(address payable dst) public payable{
        dst.send(msg.value);
    }
}
```
The return value of `send` is not checked, so if the send fails, the Ether will be locked in the contract.
If `send` is used to prevent blocking operations, consider logging the failed `send`.
    """
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Ensure that the return value of `send` is checked or logged."

    def _is_instance(self, ir: Operation) -> bool:
        return isinstance(ir, Send)
